from __future__ import annotations

from abc import ABC
from typing import Optional, Union, List, TYPE_CHECKING, Set, overload

from pymeos_cffi import *

from .tnumber import TNumber
from ..collections import *
from ..temporal import TInterpolation, Temporal, TInstant, TSequence, TSequenceSet

if TYPE_CHECKING:
    from ..boxes import TBox
    from .tfloat import TFloat


class TInt(TNumber[int, 'TInt', 'TIntInst', 'TIntSeq', 'TIntSeqSet'], ABC):
    _mobilitydb_name = 'tint'

    BaseClass = int
    _parse_function = tint_in

    # ------------------------- Constructors ----------------------------------
    @staticmethod
    def from_base_temporal(value: int, base: Temporal) -> TInt:
        """
        Returns a new temporal integer with the value `value` and the temporal
        frame of `base`.

        Args:
            value: Value of the temporal integer.
            base: Temporal object to get the temporal frame from.

        Returns:
            A new :class:`TInt` object.

        MEOS Functions:
            tint_from_base_temp
        """
        result = tint_from_base_temp(value, base._inner)
        return Temporal._factory(result)

    @staticmethod
    @overload
    def from_base_time(value: int, base: datetime) -> TIntInst:
        ...

    @staticmethod
    @overload
    def from_base_time(value: int, base: Union[TimestampSet, Period]) -> TIntSeq:
        ...

    @staticmethod
    @overload
    def from_base_time(value: int, base: PeriodSet) -> TIntSeqSet:
        ...

    @staticmethod
    def from_base_time(value: int, base: Time) -> TInt:
        """
        Returns a new temporal int with the value `value` and the temporal
        frame of `base`.

        Args:
            value: Value of the temporal int.
            base: Time object to get the temporal frame from.

        Returns:
            A new temporal int.

        MEOS Functions:
            tintinst_make, tintseq_from_base_timestampset,
            tintseq_from_base_period, tintseqset_from_base_periodset
        """
        if isinstance(base, datetime):
            return TIntInst(_inner=tintinst_make(value,
                                                 datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TIntSeq(_inner=tintseq_from_base_timestampset(value,
                                                                 base._inner))
        elif isinstance(base, Period):
            return TIntSeq(_inner=tintseq_from_base_period(value, base._inner))
        elif isinstance(base, PeriodSet):
            return TIntSeqSet(_inner=tintseqset_from_base_periodset(value,
                                                                    base._inner))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    # ------------------------- Output ----------------------------------------
    def __str__(self):
        """
        Returns a string representation of `self`.

        Returns:
            A string representation of `self`.

        MEOS Functions:
            tint_out
        """
        return tint_out(self._inner)

    def as_wkt(self):
        """
        Returns a WKT representation of `self`.

        Returns:
            A WKT representation of `self`.

        MEOS Functions:
            tint_as_wkt
        """
        return tint_out(self._inner)

    # ------------------------- Conversions ----------------------------------
    def to_tfloat(self) -> TFloat:
        """
        Returns a new temporal float with the values of `self`.

        Returns:
            A new temporal float.

        MEOS Functions:
            tint_to_tfloat
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tint_to_tfloat(self._inner))

    def to_intspan(self) -> IntSpan:
        """
        Returns value span of `self`.

        Returns:
            An :class:`IntSpan` with the value span of `self`.

        MEOS Functions:
            tnumber_to_span
        """
        return IntSpan(_inner=tnumber_to_span(self._inner))

    # ------------------------- Accessors -------------------------------------
    def value_span(self) -> IntSpan:
        """
        Returns the value span of `self`.

        Returns:
            An :class:`IntSpan` with the value span of `self`.

        MEOS Functions:
            tnumber_to_span
        """
        return self.to_intspan()

    def value_spans(self) -> IntSpanSet:
        """
        Returns the value spans of `self` taking into account gaps.

        Returns:
            An :class:`IntSpanSet` with the value spans of `self`.

        MEOS Functions:
            tnumber_valuespans
        """
        return IntSpanSet(_inner=tnumber_valuespans(self._inner))

    def start_value(self) -> int:
        """
        Returns the start value of `self`.

        Returns:
            A :class:`int` with the start value.

        MEOS Functions:
            tint_start_value
        """
        return tint_start_value(self._inner)

    def end_value(self) -> int:
        """
        Returns the end value of `self`.

        Returns:
            A :class:`int` with the end value.

        MEOS Functions:
            tint_end_value
        """
        return tint_end_value(self._inner)

    def value_set(self) -> Set[int]:
        """
        Returns the set of values of `self`.

        Returns:
            A :class:`set` with the values of `self`.

        MEOS Functions:
            tint_values
        """
        values, count = tint_values(self._inner)
        return {values[i] for i in range(count)}

    def min_value(self) -> int:
        """
        Returns the minimum value of the temporal int.

        Returns:
            A :class:`int` with the minimum value.

        MEOS Functions:
            tint_min_value
        """
        return tint_min_value(self._inner)

    def max_value(self) -> int:
        """
        Returns the maximum value of the temporal int.

        Returns:
            A :class:`int` with the maximum value.

        MEOS Functions:
            tint_max_value
        """
        return tint_max_value(self._inner)

    def value_at_timestamp(self, timestamp) -> int:
        """
        Returns the value that `self` takes at a certain moment.

        Args:
            timestamp: The moment to get the value.

        Returns:
            An :class:`int` with the value of `self` at `timestamp`.

        MEOS Functions:
            tint_value_at_timestamp
        """
        return tint_value_at_timestamp(self._inner,
                                       datetime_to_timestamptz(timestamp), True)

    # ------------------------- Ever and Always Comparisons -------------------
    def always_less(self, value: int) -> bool:
        """
        Returns whether the values of `self` are always less than `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are always less than `value`,
            `False` otherwise.

        MEOS Functions:
            tint_always_lt
        """
        return tint_always_lt(self._inner, value)

    def always_less_or_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are always less than or equal to
        `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are always less than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tint_always_le
        """
        return tint_always_le(self._inner, value)

    def always_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are always equal to `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are always equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tint_always_eq
        """
        return tint_always_eq(self._inner, value)

    def always_not_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are always not equal to `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are always not equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tint_ever_eq
        """
        return not tint_ever_eq(self._inner, value)

    def always_greater_or_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are always greater than or equal
        to `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are always greater than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tint_ever_lt
        """
        return not tint_ever_lt(self._inner, value)

    def always_greater(self, value: int) -> bool:
        """
        Returns whether the values of `self` are always greater than `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are always greater than `value`,
            `False` otherwise.

        MEOS Functions:
            tint_ever_le
        """
        return not tint_ever_le(self._inner, value)

    def ever_less(self, value: int) -> bool:
        """
        Returns whether the values of `self` are ever less than `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are ever less than `value`,
            `False` otherwise.

        MEOS Functions:
            tint_ever_lt
        """
        return tint_ever_lt(self._inner, value)

    def ever_less_or_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are ever less than or equal to
        `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are ever less than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tint_ever_le
        """
        return tint_ever_le(self._inner, value)

    def ever_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are ever equal to `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are ever equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tint_ever_eq
        """
        return tint_ever_eq(self._inner, value)

    def ever_not_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are ever not equal to `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are ever not equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tint_always_eq
        """
        return not tint_always_eq(self._inner, value)

    def ever_greater_or_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are ever greater than or equal to
        `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are ever greater than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tint_always_lt
        """
        return not tint_always_lt(self._inner, value)

    def ever_greater(self, value: int) -> bool:
        """
        Returns whether the values of `self` are ever greater than `value`.

        Args:
            value: :class:`int` to compare.

        Returns:
            `True` if the values of `self` are ever greater than `value`,
            `False` otherwise.

        MEOS Functions:
            tint_always_le
        """
        return not tint_always_le(self._inner, value)

    def never_less(self, value: int) -> bool:
        """
        Returns whether the values of `self` are never less than `value`.

        Args:
            value: :class:`int` value to compare.

        Returns:
            `True` if the values of `self` are never less than `value`,
            `False` otherwise.

        MEOS Functions:
            tint_ever_lt
        """
        return not tint_ever_lt(self._inner, value)

    def never_less_or_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are never less than or equal to
        `value`.

        Args:
            value: :class:`int` value to compare.

        Returns:
            `True` if the values of `self` are never less than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tint_ever_le
        """
        return not tint_ever_le(self._inner, value)

    def never_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are never equal to `value`.

        Args:
            value: :class:`int` value to compare.

        Returns:
            `True` if the values of `self` are never equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tint_ever_eq
        """
        return not tint_ever_eq(self._inner, value)

    def never_not_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are never not equal to `value`.

        Args:
            value: :class:`int` value to compare.

        Returns:
            `True` if the values of `self` are never not equal to `value`,
            `False` otherwise.

        MEOS Functions:
            tint_always_eq
        """
        return tint_always_eq(self._inner, value)

    def never_greater_or_equal(self, value: int) -> bool:
        """
        Returns whether the values of `self` are never greater than or equal to
        `value`.

        Args:
            value: :class:`int` value to compare.

        Returns:
            `True` if the values of `self` are never greater than or equal to
            `value`, `False` otherwise.

        MEOS Functions:
            tint_always_lt
        """
        return tint_always_lt(self._inner, value)

    def never_greater(self, value: int) -> bool:
        """
        Returns whether the values of `self` are never greater than `value`.

        Args:
            value: :class:`int` value to compare.

        Returns:
            `True` if the values of `self` are never greater than `value`,
            `False` otherwise.

        MEOS Functions:
            tint_always_le
        """
        return tint_always_le(self._inner, value)

    # ------------------------- Temporal Comparisons --------------------------
    def temporal_equal(self, other: Union[int, Temporal]) -> Temporal:
        """
        Returns the temporal equality relation between `self` and `other`.

        Args:
            other: A :class:`int` or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal equality relation.

        MEOS Functions:
            teq_tint_int, teq_temporal_temporal
        """
        if isinstance(other, int):
            result = teq_tint_int(self._inner, other)
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[int, Temporal]) -> Temporal:
        """
        Returns the temporal not equal relation between `self` and `other`.

        Args:
            other: A :class:`int` or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal not equal relation.

        MEOS Functions:
            tne_tint_int, tne_temporal_temporal
        """
        if isinstance(other, int):
            result = tne_tint_int(self._inner, other)
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    def temporal_less(self, other: Union[int, Temporal]) -> Temporal:
        """
        Returns the temporal less than relation between `self` and `other`.

        Args:
            other: A :class:`int` or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal less than relation.

        MEOS Functions:
            tlt_tint_int, tlt_temporal_temporal
        """
        if isinstance(other, int):
            result = tlt_tint_int(self._inner, other)
        else:
            return super().temporal_less(other)
        return Temporal._factory(result)

    def temporal_less_or_equal(self, other: Union[int, Temporal]) -> Temporal:
        """
        Returns the temporal less or equal relation between `self` and `other`.

        Args:
            other: A :class:`int` or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal less or equal
            relation.

        MEOS Functions:
            tle_tint_int, tle_temporal_temporal
        """
        if isinstance(other, int):
            result = tle_tint_int(self._inner, other)
        else:
            return super().temporal_less_or_equal(other)
        return Temporal._factory(result)

    def temporal_greater_or_equal(self, other: Union[int, Temporal]) -> Temporal:
        """
        Returns the temporal greater or equal relation between `self` and `other`.

        Args:
            other: A :class:`int` or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal greater or equal
            relation.

        MEOS Functions:
            tge_tint_int, tge_temporal_temporal
        """
        if isinstance(other, int):
            result = tge_tint_int(self._inner, other)
        else:
            return super().temporal_greater_or_equal(other)
        return Temporal._factory(result)

    def temporal_greater(self, other: Union[int, Temporal]) -> Temporal:
        """
        Returns the temporal greater than relation between `self` and `other`.

        Args:
            other: A :class:`int` or temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal greater than
            relation.

        MEOS Functions:
            tgt_tint_int, tgt_temporal_temporal
        """
        if isinstance(other, int):
            result = tgt_tint_int(self._inner, other)
        else:
            return super().temporal_greater(other)
        return Temporal._factory(result)

    # ------------------------- Restrictions ----------------------------------
    def at(self, other: Union[int, float, IntSet, FloatSet, IntSpan, FloatSpan, \
            IntSpanSet, FloatSpanSet, TBox, Time]) -> Temporal:
        """
        Returns a new temporal int with th  e values of `self` restricted to
        the time or value `other`.

        Args:
            other: Time or value to restrict to.

        Returns:
            A new temporal int.

        MEOS Functions:
            tint_at_value, temporal_at_values, tnumber_at_span, tnumber_at_spanset,
            temporal_at_timestamp, temporal_at_timestampset, temporal_at_period,
            temporal_at_periodset
        """
        if isinstance(other, int) or isinstance(other, float):
            result = tint_at_value(self._inner, int(other))
        elif isinstance(other, FloatSet):
            return super().at(other.to_intset())
        elif isinstance(other, FloatSpan):
            return super().at(other.to_intspan())
        elif isinstance(other, FloatSpanSet):
            return super().at(other.to_intspanset())
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[int, float, IntSet, FloatSet, IntSpan, FloatSpan, \
            IntSpanSet, FloatSpanSet, TBox, Time]) -> Temporal:
        """
        Returns a new temporal int with the values of `self` restricted to the
        complement of the time or value `other`.

        Args:
            other: Time or value to restrict to the complement of.

        Returns:
            A new temporal int.

        MEOS Functions:
            tint_minus_value, temporal_minus_values, tnumber_minus_span, tnumber_minus_spanset,
            temporal_minus_timestamp, temporal_minus_timestampset,
            temporal_minus_period, temporal_minus_periodset
        """
        if isinstance(other, int) or isinstance(other, float):
            result = tint_minus_value(self._inner, int(other))
        elif isinstance(other, FloatSet):
            return super().minus(other.to_intset())
        elif isinstance(other, FloatSpan):
            return super().minus(other.to_intspan())
        elif isinstance(other, FloatSpanSet):
            return super().minus(other.to_intspanset())
        else:
            return super().minus(other)
        return Temporal._factory(result)

    # ------------------------- Distance --------------------------------------
    def nearest_approach_distance(self,
                                  other: Union[int, float, TNumber, TBox]) -> float:
        """
        Returns the nearest approach distance between `self` and `other`.

        Args:
            other: A :class:`int`, :class:`float`, :class:`TNumber` or 
            :class:`TBox` to compare to `self`.

        Returns:
            A :class:`float` with the nearest approach distance between `self`
            and `other`.

        MEOS Functions:
            nad_tint_int, nad_tint_tint, nad_tfloat_float, nad_tfloat_tfloat,
            nad_tnumber_tbox
        """
        if isinstance(other, int):
            return nad_tint_int(self._inner, other)
        elif isinstance(other, TInt):
            return nad_tint_tint(self._inner, other._inner)
        else:
            return super().nearest_approach_distance(other)

    # ------------------------- Split Operations ------------------------------
    def value_split(self, size: int, start: Optional[int] = 0) -> List[TInt]:
        """
        Splits `self` into fragments with respect to value buckets

        Args:
            start: Start value of the first value bucket.
            size: Size of the value buckets.

        Returns:
            A list of temporal ints.

        MEOS Functions:
            tint_value_split
        """
        fragments, values, count = tint_value_split(self._inner, size, start)
        return [Temporal._factory(fragments[i]) for i in range(count)]

    def value_time_split(self, value_size: int,
                         duration: Union[str, timedelta],
                         value_start: Optional[int] = 0,
                         time_start: Optional[Union[str, datetime]] = None) -> \
            List[TInt]:
        """
        Splits `self` into fragments with respect to value and period buckets.

        Args:
            value_size: Size of the value buckets.
            duration: Duration of the period buckets.
            value_start: Start value of the first value bucket. If None, the
                start value used by default is 0
            time_start: Start time of the first period bucket. If None, the
                start time used by default is Monday, January 3, 2000.

        Returns:
            A list of temporal integers.

        MEOS Functions:
            tint_value_time_split
        """
        if time_start is None:
            st = pg_timestamptz_in('2000-01-03', -1)
        else:
            st = datetime_to_timestamptz(time_start) \
                if isinstance(time_start, datetime) \
                else pg_timestamptz_in(time_start, -1)
        dt = timedelta_to_interval(duration) \
            if isinstance(duration, timedelta) \
            else pg_interval_in(duration, -1)
        tiles, _, _, count = tint_value_time_split(self._inner, value_size, dt,
                                                   value_start, st)
        return [Temporal._factory(tiles[i]) for i in range(count)]

    # ------------------------- Database Operations ---------------------------
    @staticmethod
    def read_from_cursor(value, _=None):
        """
        Reads a :class:`TInt` from a database cursor. Used when automatically
        loading objects from the database.
        Users should use the class constructor instead.
        """
        if not value:
            return None
        if value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TIntInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TIntSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TIntSeqSet(string=value)
            else:
                return TIntSeq(string=value)
        raise Exception("ERROR: Could not parse temporal integer value")


class TIntInst(TInstant[int, 'TInt', 'TIntInst', 'TIntSeq', 'TIntSeqSet'], TInt):
    """
    Class for representing temporal integers at a single instant.
    """
    _make_function = tintinst_make
    _cast_function = int

    def __init__(self, string: Optional[str] = None, *,
                 value: Optional[Union[str, int]] = None,
                 timestamp: Optional[Union[str, datetime]] = None, _inner=None):
        super().__init__(string=string, value=value, timestamp=timestamp,
                         _inner=_inner)


class TIntSeq(TSequence[int, 'TInt', 'TIntInst', 'TIntSeq', 'TIntSeqSet'], TInt):
    """
    Class for representing temporal integers over a period of time.
    """
    ComponentClass = TIntInst

    def __init__(self, string: Optional[str] = None, *,
                 instant_list: Optional[List[Union[str, TIntInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False,
                 interpolation: TInterpolation = TInterpolation.STEPWISE,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list,
                         lower_inc=lower_inc, upper_inc=upper_inc,
                         interpolation=interpolation, normalize=normalize, _inner=_inner)


class TIntSeqSet(TSequenceSet[int, 'TInt', 'TIntInst', 'TIntSeq', 'TIntSeqSet'],
                 TInt):
    """
    Class for representing temporal integers over a period of time with gaps.
    """
    ComponentClass = TIntSeq

    def __init__(self, string: Optional[str] = None, *,
                 sequence_list: Optional[List[Union[str, TIntSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list,
                         normalize=normalize, _inner=_inner)



from __future__ import annotations

from typing import Union, overload, Optional, TYPE_CHECKING

from pymeos_cffi import intspan_in, intspan_lower, intspan_upper, \
    intspan_shift_scale, contains_intspan_int, adjacent_intspan_int, \
    span_width, int_to_intspan, span_eq, left_intspan_int, \
    overleft_intspan_int, right_intspan_int, overright_intspan_int, \
    intersection_intspan_int, intersection_span_span, intersection_spanset_span, \
    minus_intspan_int, minus_span_span, minus_spanset_span, union_intspan_int, \
    union_span_span, union_spanset_span, intspan_out, intspan_make, \
    distance_intspan_int, intspan_floatspan

from .. import Span

if TYPE_CHECKING:
    from .intspanset import IntSpanSet
    from .floatspan import FloatSpan


class IntSpan(Span[int]):
    """
    Class for representing sets of contiguous integer values between a lower and
    an upper bound. The bounds may be inclusive or not.

    ``IntSpan`` objects can be created with a single argument of type string
    as in MobilityDB.

        >>> IntSpan('(2, 5]')

    Another possibility is to provide the ``lower`` and ``upper`` named parameters (of type str or int), and
    optionally indicate whether the bounds are inclusive or exclusive (by default, the lower bound is inclusive and the
    upper is exclusive):

        >>> IntSpan(lower=2, upper=5)
        >>> IntSpan(lower=2, upper=5, lower_inc=False, upper_inc=True)
        >>> IntSpan(lower='2', upper='5', upper_inc=True)
    """

    __slots__ = ['_inner']

    _mobilitydb_name = 'intspan'

    _parse_function = intspan_in
    _parse_value_function = int
    _make_function = intspan_make

    # ------------------------- Output ----------------------------------------
    def __str__(self):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            intspan_out
        """
        return intspan_out(self._inner)

    # ------------------------- Conversions -----------------------------------

    def to_spanset(self) -> IntSpanSet:
        """
        Returns a spanset containing ``self``.

        Returns:
            A new :class:`IntSpanSet` instance

        MEOS Functions:
            span_to_spanset
        """
        from .intspanset import IntSpanSet
        return IntSpanSet(_inner=super().to_spanset())

    def to_floatspan(self) -> FloatSpan:
        """
        Converts ``self`` to a :class:`FloatSpan` instance.

        Returns:
            A new :class:`FloatSpan` instance

        MEOS Functions:
            intspan_floatspan
        """
        from .floatspan import FloatSpan
        return FloatSpan(_inner=intspan_floatspan(self._inner))

    # ------------------------- Accessors -------------------------------------
    def lower(self) -> int:
        """
        Returns the lower bound of ``self``.

        Returns:
            The lower bound of the span as a :class:`int`

        MEOS Functions:
            period_lower
        """

        return intspan_lower(self._inner)

    def upper(self) -> int:
        """
        Returns the upper bound of ``self``.

        Returns:
            The upper bound of the span as a :class:`int`

        MEOS Functions:
            period_upper
        """
        return intspan_upper(self._inner)

    def width(self) -> float:
        """
        Returns the width of ``self``.

        Returns:
            Returns a `float` representing the width of the span

        MEOS Functions:
            span_width
        """
        return span_width(self._inner)

    # ------------------------- Transformations -------------------------------
    def shift(self, delta: int) -> IntSpan:
        """
        Return a new ``IntSpan`` with the lower and upper bounds shifted by ``delta``.

        Args:
            delta: The value to shift by

        Returns:
            A new ``IntSpan`` instance

        MEOS Functions:
            intspan_shift_scale
        """
        return self.shift_scale(delta, None)

    def scale(self, width: int) -> IntSpan:
        """
        Return a new ``IntSpan`` with the lower and upper bounds scaled so that
        the width is ``width``.

        Args:
            width: The new width

        Returns:
            A new ``IntSpan`` instance

        MEOS Functions:
            intspan_shift_scale
        """
        return self.shift_scale(None, width)

    def shift_scale(self, delta: Optional[int], width: Optional[int]) -> IntSpan:
        """
        Return a new ``IntSpan`` with the lower and upper bounds shifted by
        ``delta`` and scaled so that the width is ``width``.

        Args:
            delta: The value to shift by
            width: The new width

        Returns:
            A new ``IntSpan`` instance

        MEOS Functions:
            intspan_shift_scale
        """
        d = delta if delta is not None else 0
        w = width if width is not None else 0
        modified = intspan_shift_scale(self._inner, d, w, delta is not None,
                                       width is not None)
        return IntSpan(_inner=modified)

    # ------------------------- Topological Operations --------------------------------

    def is_adjacent(self, other: Union[int, IntSpan, IntSpanSet]) -> bool:
        """
        Returns whether ``self`` is adjacent to ``other``. That is, they share
        a bound but only one of them contains it.

        Args:
            other: object to compare with

        Returns:
            True if adjacent, False otherwise

        MEOS Functions:
            adjacent_span_span, adjacent_span_spanset, adjacent_intspan_int
        """
        if isinstance(other, int):
            return adjacent_intspan_int(self._inner, other)
        else:
            return super().is_adjacent(other)

    def contains(self, content: Union[int, IntSpan, IntSpanSet]) -> bool:
        """
        Returns whether ``self`` contains ``content``.

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_span_span, contains_intspan_int
        """
        if isinstance(content, int):
            return contains_intspan_int(self._inner, content)
        else:
            return super().contains(content)

    def is_same(self, other: Union[int, IntSpan, IntSpanSet]) -> bool:
        """
        Returns whether ``self`` and the bounding period of ``other`` is the
        same.

        Args:
            other: object to compare with

        Returns:
            True if equal, False otherwise

        MEOS Functions:
            same_period_temporal
        """
        if isinstance(other, int):
            return span_eq(self._inner, int_to_intspan(other))
        else:
            return super().is_same(other)

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other: Union[int, IntSpan, IntSpanSet]) -> bool:
        """
        Returns whether ``self`` is strictly left ``other``. That is,
        ``self`` ends before ``other`` starts.

        Args:
            other: object to compare with

        Returns:
            True if left, False otherwise

        MEOS Functions:
            left_span_span, left_span_spanset, left_intspan_int
        """
        if isinstance(other, int):
            return left_intspan_int(self._inner, other)
        else:
            return super().is_left(other)

    def is_over_or_left(self, other: Union[int, IntSpan, IntSpanSet]) -> bool:
        """
        Returns whether ``self`` is left ``other`` allowing overlap. That is,
        ``self`` ends before ``other`` ends (or at the same value).

        Args:
            other: object to compare with

        Returns:
            True if over or left, False otherwise

        MEOS Functions:
            overleft_span_span, overleft_span_spanset, overleft_intspan_int
        """
        if isinstance(other, int):
            return overleft_intspan_int(self._inner, other)
        else:
            return super().is_over_or_left(other)

    def is_right(self, other: Union[int, IntSpan, IntSpanSet]) -> bool:
        """
        Returns whether ``self`` is strictly right ``other``. That is, ``self``
        starts after ``other`` ends.

        Args:
            other: object to compare with

        Returns:
            True if right, False otherwise

        MEOS Functions:
            right_span_span, right_span_spanset, right_intspan_int
        """
        if isinstance(other, int):
            return right_intspan_int(self._inner, other)
        else:
            return super().is_right(other)

    def is_over_or_right(self, other: Union[int, IntSpan, IntSpanSet]) -> bool:
        """
        Returns whether ``self`` is right ``other`` allowing overlap. That is,
        ``self`` starts after ``other`` starts (or at the same value).

        Args:
            other: object to compare with

        Returns:
            True if overlapping or after, False otherwise

        MEOS Functions:
            overright_span_span, overright_span_spanset, overright_intspan_int
        """
        if isinstance(other, int):
            return overright_intspan_int(self._inner, other)
        else:
            return super().is_over_or_right(other)

    # ------------------------- Distance Operations ---------------------------
    def distance(self, other: Union[int, IntSpan, IntSpanSet]) -> float:
        """
        Returns the distance between ``self`` and ``other``.

        Args:
            other: object to compare with

        Returns:
            A float value

        MEOS Functions:
            distance_span_span, distance_span_spanset, distance_intspan_int,
        """
        if isinstance(other, int):
            return distance_intspan_int(self._inner, other)
        else:
            return super().distance(other)

    # ------------------------- Set Operations --------------------------------
    @overload
    def intersection(self, other: int) -> Optional[int]:
        ...

    @overload
    def intersection(self, other: IntSpan) -> Optional[IntSpan]:
        ...

    @overload
    def intersection(self, other: IntSpanSet) -> Optional[IntSpanSet]:
        ...

    def intersection(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: object to intersect with

        Returns:
            A :class:`Collection[int]` instance. The actual class depends on
            ``other``.

        MEOS Functions:
            intersection_span_span, intersection_spanset_span,
            intersection_intspan_int
        """
        from .intspanset import IntSpanSet
        if isinstance(other, int):
            return intersection_intspan_int(self._inner, other)
        elif isinstance(other, IntSpan):
            result = intersection_span_span(self._inner, other._inner)
            return IntSpan(_inner=result) if result is not None else None
        elif isinstance(other, IntSpanSet):
            result = intersection_spanset_span(other._inner, self._inner)
            return IntSpanSet(_inner=result) if result is not None else None
        else:
            super().intersection(other)

    def minus(self, other: Union[int, IntSpan, IntSpanSet]) -> IntSpanSet:
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: object to diff with

        Returns:
            A :class:`IntSpanSet` instance.

        MEOS Functions:
            minus_span_span, minus_spanset_span, minus_intspan_int
        """
        from .intspanset import IntSpanSet
        if isinstance(other, int):
            result = minus_intspan_int(self._inner, other)
        elif isinstance(other, IntSpan):
            result = minus_span_span(self._inner, other._inner)
        elif isinstance(other, IntSpanSet):
            result = minus_spanset_span(other._inner, self._inner)
        else:
            result = super().minus(other)
        return IntSpanSet(_inner=result) if result is not None else None

    def union(self, other: Union[int, IntSpan, IntSpanSet]) -> IntSpanSet:
        """
        Returns the union of ``self`` and ``other``.

        Args:
            other: object to merge with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
            union_spanset_span, union_span_span, union_intspan_int
        """
        from .intspanset import IntSpanSet
        if isinstance(other, int):
            result = union_intspan_int(self._inner, other)
        elif isinstance(other, IntSpan):
            result = union_span_span(self._inner, other._inner)
        elif isinstance(other, IntSpanSet):
            result = union_spanset_span(other._inner, self._inner)
        else:
            result = super().union(other)
        return IntSpanSet(_inner=result) if result is not None else None



from .intset import IntSet
from .intspan import IntSpan
from .intspanset import IntSpanSet
from .floatset import FloatSet
from .floatspan import FloatSpan
from .floatspanset import FloatSpanSet

__all__ = [
    'IntSet', 'IntSpan', 'IntSpanSet',
    'FloatSet', 'FloatSpan', 'FloatSpanSet'
]



from __future__ import annotations

from abc import ABC
from typing import Optional, Union, List, Any, TypeVar, Type

from pymeos_cffi import *

from .interpolation import TInterpolation
from .temporal import Temporal

TBase = TypeVar('TBase')
TG = TypeVar('TG', bound='Temporal[Any]')
TI = TypeVar('TI', bound='TInstant[Any]')
TS = TypeVar('TS', bound='TSequence[Any]')
TSS = TypeVar('TSS', bound='TSequenceSet[Any]')
Self = TypeVar('Self', bound='TSequence[Any]')


class TSequence(Temporal[TBase, TG, TI, TS, TSS], ABC):
    """
    Base class for temporal sequence types, i.e. temporal values that are
    defined over a continuous period of time.
    """

    # ------------------------- Constructors ----------------------------------
    def __init__(self, string: Optional[str] = None, *,
                 instant_list: Optional[List[Union[str, Any]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False,
                 interpolation: TInterpolation = TInterpolation.LINEAR,
                 normalize: bool = True, _inner=None):
        assert (_inner is not None) or ((string is not None) != (instant_list is not None)), \
            "Either string must be not None or instant_list must be not"
        if _inner is not None:
            self._inner = as_tsequence(_inner)
        elif string is not None:
            self._inner = as_tsequence(self.__class__._parse_function(string))
        else:
            self._instants = [x._inner if isinstance(x, self.ComponentClass) \
                                  else self.__class__._parse_function(x) for x in instant_list]
            count = len(self._instants)
            self._inner = tsequence_make(self._instants, count, lower_inc,
                                         upper_inc, interpolation.value, normalize)

    @classmethod
    def from_instants(cls: Type[Self],
                      instant_list: Optional[List[Union[str, Any]]],
                      lower_inc: bool = True, upper_inc: bool = False,
                      interpolation: TInterpolation = TInterpolation.LINEAR,
                      normalize: bool = True) -> Self:
        """
        Create a temporal sequence from a list of instants.

        Args:
            instant_list: List of instants
            lower_inc: Whether the lower bound is inclusive
            upper_inc: Whether the upper bound is inclusive
            interpolation: Interpolation method
            normalize: Whether to normalize the sequence

        Returns:
            A new temporal sequence.
        """
        return cls(instant_list=instant_list, lower_inc=lower_inc,
                   upper_inc=upper_inc, interpolation=interpolation,
                   normalize=normalize)

    # ------------------------- Accessors -------------------------------------
    def lower_inc(self) -> bool:
        """
        Returns whether the lower bound is inclusive.

        Returns:
            `True` if the lower bound is inclusive, `False` otherwise.
        """
        return self._inner.period.lower_inc

    def upper_inc(self) -> bool:
        """
        Returns whether the upper bound is inclusive.

        Returns:
            `True` if the upper bound is inclusive, `False` otherwise.
        """
        return self._inner.period.upper_inc

    # ------------------------- Plot Operations -------------------------------
    def plot(self, *args, **kwargs):
        """
        Plot the temporal sequence.

        See Also:
            :meth:`pymeos.plotters.TemporalSequencePlotter.plot`
        """
        from ..plotters import TemporalSequencePlotter
        return TemporalSequencePlotter.plot(self, *args, **kwargs)



from __future__ import annotations

from datetime import timedelta, datetime
from typing import Optional, Union, List, overload, get_args
from typing import TYPE_CHECKING

from pymeos_cffi import *

from .time_collection import TimeCollection

if TYPE_CHECKING:
    from ...temporal import Temporal
    from ...boxes import Box
    from .period import Period
    from .timestampset import TimestampSet
    from .time import Time

from ..base.spanset import SpanSet


class PeriodSet(SpanSet[datetime], TimeCollection):
    """
    Class for representing lists of disjoint periods.

    ``PeriodSet`` objects can be created with a single argument of type string
    as in MobilityDB.

        >>> PeriodSet(string='{[2019-09-08 00:00:00+01, 2019-09-10 00:00:00+01], [2019-09-11 00:00:00+01, 2019-09-12 00:00:00+01]}')

    Another possibility is to give a list specifying the composing
    periods, which can be instances  of ``str`` or ``Period``. The composing
    periods must be given in increasing order.

        >>> PeriodSet(span_list=['[2019-09-08 00:00:00+01, 2019-09-10 00:00:00+01]', '[2019-09-11 00:00:00+01, 2019-09-12 00:00:00+01]'])
        >>> PeriodSet(span_list=[Period('[2019-09-08 00:00:00+01, 2019-09-10 00:00:00+01]'), Period('[2019-09-11 00:00:00+01, 2019-09-12 00:00:00+01]')])

    """

    __slots__ = ['_inner']

    _mobilitydb_name = 'tstzspanset'

    _parse_function = periodset_in
    _parse_value_function = lambda period: period_in(period)[0] if isinstance(period, str) else period._inner[0]

    # ------------------------- Output ----------------------------------------
    def __str__(self):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            periodset_out
        """
        return periodset_out(self._inner)

    # ------------------------- Conversions -----------------------------------

    def to_span(self) -> Period:
        """
        Returns a period that encompasses ``self``.

        Returns:
            A new :class:`Period` instance

        MEOS Functions:
            spanset_span
        """
        from .period import Period
        return Period(_inner=super().to_span())

    def to_period(self) -> Period:
        """
        Returns a period that encompasses ``self``.

        Returns:
            A new :class:`Period` instance

        MEOS Functions:
            spanset_span
        """
        return self.to_span()

    # ------------------------- Accessors -------------------------------------
    def duration(self, ignore_gaps: Optional[bool] = False) -> timedelta:
        """
        Returns the duration of the periodset. By default, i.e., when the
        second argument is False, the function takes into account the gaps within,
        i.e., returns the sum of the durations of the periods within.
        Otherwise, the function returns the duration of the periodset ignoring
        any gap, i.e., the duration from the lower bound of the first period to
        the upper bound of the last period.

        Parameters:
            ignore_gaps: Whether to take into account potential time gaps in
            the periodset.

        Returns:
            A :class:`datetime.timedelta` instance representing the duration of
            the periodset

        MEOS Functions:
            periodset_duration
        """
        return interval_to_timedelta(periodset_duration(self._inner, ignore_gaps))

    def num_timestamps(self) -> int:
        """
        Returns the number of timestamps in ``self``.
        Returns:
            An :class:`int`

        MEOS Functions:
            periodset_num_timestamps
        """
        return periodset_num_timestamps(self._inner)

    def start_timestamp(self) -> datetime:
        """
        Returns the first timestamp in ``self``.
        Returns:
            A :class:`datetime` instance

        MEOS Functions:
            periodset_start_timestamp
        """
        return timestamptz_to_datetime(periodset_start_timestamp(self._inner))

    def end_timestamp(self) -> datetime:
        """
        Returns the last timestamp in ``self``.
        Returns:
            A :class:`datetime` instance

        MEOS Functions:
            periodset_end_timestamp
        """
        return timestamptz_to_datetime(periodset_end_timestamp(self._inner))

    def timestamp_n(self, n: int) -> datetime:
        """
        Returns the n-th timestamp in ``self``.
        Returns:
            A :class:`datetime` instance

        MEOS Functions:
            periodset_timestamp_n
        """
        if n < 0 or n >= self.num_timestamps():
            raise IndexError(f'Index {n} out of bounds')
        return timestamptz_to_datetime(periodset_timestamp_n(self._inner, n + 1))

    def timestamps(self) -> List[datetime]:
        """
        Returns the list of distinct timestamps in ``self``.
        Returns:
            A :class:`list[datetime]` instance

        MEOS Functions:
            periodset_timestamps
        """
        ts, count = periodset_timestamps(self._inner)
        return [timestamptz_to_datetime(ts[i]) for i in range(count)]

    def num_periods(self) -> int:
        """
        Returns the number of periods in ``self``.
        Returns:
            An :class:`int`

        MEOS Functions:
            spanset_num_spans
        """
        return self.num_spans()

    def start_span(self) -> Period:
        """
        Returns the first period in ``self``.
        Returns:
            A :class:`Period` instance

        MEOS Functions:
            spanset_start_span
        """
        from .period import Period
        return Period(_inner=super().start_span())

    def start_period(self) -> Period:
        """
        Returns the first period in ``self``.
        Returns:
            A :class:`Period` instance

        MEOS Functions:
            spanset_start_span
        """
        return self.start_span()

    def end_span(self) -> Period:
        """
        Returns the last period in ``self``.
        Returns:
            A :class:`Period` instance

        MEOS Functions:
            spanset_end_span
        """
        from .period import Period
        return Period(_inner=super().end_span())

    def end_period(self) -> Period:
        """
        Returns the last period in ``self``.
        Returns:
            A :class:`Period` instance

        MEOS Functions:
            spanset_end_span
        """
        return self.end_span()

    def span_n(self, n: int) -> Period:
        """
        Returns the n-th period in ``self``.
        Returns:
            A :class:`Period` instance

        MEOS Functions:
            spanset_span_n
        """
        from .period import Period
        return Period(_inner=super().span_n(n))

    def period_n(self, n: int) -> Period:
        """
        Returns the n-th period in ``self``.
        Returns:
            A :class:`Period` instance

        MEOS Functions:
            spanset_span_n
        """
        return self.span_n(n)

    def spans(self) -> List[Period]:
        """
        Returns the list of periods in ``self``.
        Returns:
            A :class:`list[Period]` instance

        MEOS Functions:
            spanset_spans
        """
        from .period import Period
        ps = super().spans()
        return [Period(_inner=ps[i]) for i in range(self.num_spans())]

    def periods(self) -> List[Period]:
        """
        Returns the list of periods in ``self``.
        Returns:
            A :class:`list[Period]` instance

        MEOS Functions:
            spanset_spans
        """
        return self.spans()

    # ------------------------- Transformations -------------------------------
    def shift(self, delta: timedelta) -> PeriodSet:
        """
        Returns a new periodset that is the result of shifting ``self`` by
        ``delta``

        Examples:
            >>> Period('[2000-01-01, 2000-01-10]').shift(timedelta(days=2))
            >>> 'Period([2000-01-03 00:00:00+01, 2000-01-12 00:00:00+01])'

        Args:
            delta: :class:`datetime.timedelta` instance to shift

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            periodset_shift_scale
        """
        return self.shift_scale(shift=delta)

    def scale(self, duration: timedelta) -> PeriodSet:
        """
        Returns a new periodset that starts as ``self`` but has duration
        ``duration``

        Examples:
            >>> Period('[2000-01-01, 2000-01-10]').scale(timedelta(days=2))
            >>> 'Period([2000-01-01 00:00:00+01, 2000-01-03 00:00:00+01])'

        Args:
            duration: :class:`datetime.timedelta` instance representing the
            duration of the new period

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            periodset_shift_scale
        """
        return self.shift_scale(duration=duration)

    def shift_scale(self, shift: Optional[timedelta] = None,
        duration: Optional[timedelta] = None) -> PeriodSet:
        """
        Returns a new periodset that starts at ``self`` shifted by ``shift``
        and has duration ``duration``

        Examples:
            >>> Period('[2000-01-01, 2000-01-10]').shift_scale(shift=timedelta(days=2), duration=timedelta(days=4))
            >>> 'Period([2000-01-03 00:00:00+01, 2000-01-07 00:00:00+01])'

        Args:
            shift: :class:`datetime.timedelta` instance to shift
            duration: :class:`datetime.timedelta` instance representing the
            duration of the new period

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            periodset_shift_scale
        """
        assert shift is not None or duration is not None, 'shift and scale deltas must not be both None'
        ps = periodset_shift_scale(
            self._inner,
            timedelta_to_interval(shift) if shift else None,
            timedelta_to_interval(duration) if duration else None
        )
        return PeriodSet(_inner=ps)

    # ------------------------- Topological Operations ------------------------
    def is_adjacent(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is temporally adjacent to ``other``. That is,
        they share a bound but only one of them contains it.

        Examples:
            >>> PeriodSet('{[2012-01-01, 2012-01-02)}').is_adjacent(PeriodSet('{[2012-01-02, 2012-01-03]}'))
            >>> True
            >>> PeriodSet('{[2012-01-01, 2012-01-02]}').is_adjacent(PeriodSet('{[2012-01-02, 2012-01-03]}'))
            >>> False  # Both contain bound
            >>> PeriodSet('{[2012-01-01, 2012-01-02)}').is_adjacent(PeriodSet('{[(2012-01-02, 2012-01-03]]}'))
            >>> False  # Neither contain bound

        Args:
            other: temporal object to compare with

        Returns:
            True if adjacent, False otherwise

        MEOS Functions:
            adjacent_spanset_span, adjacent_spanset_spanset, 
            adjacent_periodset_timestamp, adjacent_periodset_timestampset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return adjacent_periodset_timestamp(self._inner,
                datetime_to_timestamptz(other))
        elif isinstance(other, Temporal):
            return self.is_adjacent(other.period())
        elif isinstance(other, get_args(Box)):
            return self.is_adjacent(other.to_period())
        else:
            return super().is_adjacent(other)

    def is_contained_in(self, container: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is temporally contained in ``container``.

        Examples:
            >>> PeriodSet('{[2012-01-02, 2012-01-03]}').is_contained_in(Period('{[2012-01-01, 2012-01-04]}'))
            >>> True
            >>> PeriodSet('{(2012-01-01, 2012-01-02)}').is_contained_in(Period('{[2012-01-01, 2012-01-02]}'))
            >>> True
            >>> PeriodSet('{[2012-01-01, 2012-01-02]}').is_contained_in(Period('{(2012-01-01, 2012-01-02)}'))
            >>> False

        Args:
            container: temporal object to compare with

        Returns:
            True if contained, False otherwise

        MEOS Functions:
            contained_spanset_span, contained_spanset_spanset, 
            contained_periodset_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(container, datetime):
            return contained_spanset_span(self._inner, 
                timestamp_to_period(datetime_to_timestamptz(container)))
        elif isinstance(container, Temporal):
            return self.is_contained_in(container.period())
        elif isinstance(container, get_args(Box)):
            return self.is_contained_in(container.to_period())
        else:
            return super().is_contained_in(container)

    def contains(self, content: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` temporally contains ``content``.

        Examples:
            >>> PeriodSet('{[2012-01-01, 2012-01-04]}').contains(PeriodSet('{[2012-01-02, 2012-01-03]}'))
            >>> True
            >>> PeriodSet('{[2012-01-01, 2012-01-02]}').contains(PeriodSet('{(2012-01-01, 2012-01-02)}'))
            >>> True
            >>> PeriodSet('{(2012-01-01, 2012-01-02)}').contains(PeriodSet('{[2012-01-01, 2012-01-02]}'))
            >>> False

        Args:
            content: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_spanset_span, contains_spanset_spanset,
            contains_periodset_timestamp
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(content, datetime):
            return contains_periodset_timestamp(self._inner,
                datetime_to_timestamptz(content))
        elif isinstance(content, Temporal):
            return self.contains(content.period())
        elif isinstance(content, get_args(Box)):
            return self.contains(content.to_period())
        else:
            return super().contains(content)

    def __contains__(self, item):
        """
        Returns whether ``self`` temporally contains ``content``.

        Examples:
            >>> PeriodSet('{[2012-01-01, 2012-01-04]}').contains(PeriodSet('{[2012-01-02, 2012-01-03]}'))
            >>> True
            >>> PeriodSet('{[2012-01-01, 2012-01-02]}').contains(PeriodSet('{(2012-01-01, 2012-01-02)}'))
            >>> True
            >>> PeriodSet('{(2012-01-01, 2012-01-02)}').contains(PeriodSet('{[2012-01-01, 2012-01-02]}'))
            >>> False

        Args:
            item: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_spanset_span, contains_spanset_spanset, 
            contains_periodset_timestamp, contains_periodset_timestampset,
            contains_periodset_temporal
        """
        return self.contains(item)

    def overlaps(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` temporally overlaps ``other``. That is, both
        share at least an instant

        Examples:
            >>> PeriodSet('{[2012-01-01, 2012-01-02]}').overlaps(PeriodSet('{[2012-01-02, 2012-01-03]}'))
            >>> True
            >>> PeriodSet('{[2012-01-01, 2012-01-02)}').overlaps(PeriodSet('{[2012-01-02, 2012-01-03]}'))
            >>> False
            >>> PeriodSet('{[2012-01-01, 2012-01-02)}').overlaps(PeriodSet('{(2012-01-02, 2012-01-03]}'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if overlaps, False otherwise

        MEOS Functions:
            overlaps_spanset_span, overlaps_spanset_spanset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overlaps_spanset_span(self._inner,
                timestamp_to_period(datetime_to_timestamptz(other)))
        elif isinstance(other, Temporal):
            return self.overlaps(other.period())
        elif isinstance(other, get_args(Box)):
            return self.overlaps(other.to_period())
        else:
            return super().overlaps(other)

    def is_same(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether the bounding period of `self` is the same as the
        bounding period of `other`.

        Args:
            other: A time or temporal object to compare to `self`.

        Returns:
            True if same, False otherwise.

        See Also:
            :meth:`Period.is_same`
        """
        return self.to_period().is_same(other)

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is strictly before ``other``. That is,
        ``self`` ends before ``other`` starts.

        Examples:
            >>> PeriodSet('{[2012-01-01, 2012-01-02)}').is_left(PeriodSet('{[2012-01-02, 2012-01-03]}'))
            >>> True
            >>> PeriodSet('{[2012-01-01, 2012-01-02)}').is_left(PeriodSet('{(2012-01-02, 2012-01-03]}'))
            >>> True
            >>> PeriodSet('{[2012-01-01, 2012-01-02]}').is_left(PeriodSet('{[2012-01-02, 2012-01-03]}'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
        before_periodset_timestamp, left_spanset_span, left_spanset_spanset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return before_periodset_timestamp(self._inner, 
                datetime_to_timestamptz(other))
        elif isinstance(other, Temporal):
            return self.is_left(other.period())
        elif isinstance(other, get_args(Box)):
            return self.is_left(other.to_period())
        else:
            return super().is_left(other)

    def is_over_or_left(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is before ``other`` allowing overlap. That is,
        ``self`` ends before ``other`` ends (or at the same time).

        Examples:
            >>> PeriodSet('{[2012-01-01, 2012-01-02)}').is_over_or_left(PeriodSet('{[2012-01-02, 2012-01-03]}'))
            >>> True
            >>> PeriodSet('{[2012-01-01, 2012-01-02]}').is_over_or_left(PeriodSet('{[2012-01-02, 2012-01-03]}'))
            >>> True
            >>> PeriodSet('{[2012-01-03, 2012-01-05]}').is_over_or_left(PeriodSet('{[2012-01-01, 2012-01-04]}'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            overleft_spanset_span, overleft_spanset_spanset, 
            overbefore_periodset_timestamp, overbefore_periodset_timestampset,
            overbefore_periodset_temporal
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overbefore_periodset_timestamp(self._inner,
                datetime_to_timestamptz(other))
        elif isinstance(other, Temporal):
            return self.is_over_or_left(other.period())
        elif isinstance(other, get_args(Box)):
            return self.is_over_or_left(other.to_period())
        else:
            return super().is_over_or_left(other)

    def is_over_or_right(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is after ``other`` allowing overlap. That is,
        ``self`` starts after ``other`` starts (or at the same time).

        Examples:
            >>> PeriodSet('{[2012-01-02, 2012-01-03]}').is_over_or_right(PeriodSet('{[2012-01-01, 2012-01-02)}'))
            >>> True
            >>> PeriodSet('{[2012-01-02, 2012-01-03]}').is_over_or_right(PeriodSet('{[2012-01-01, 2012-01-02]}'))
            >>> True
            >>> PeriodSet('{[2012-01-02, 2012-01-03]}').is_over_or_right(PeriodSet('{[2012-01-01, 2012-01-03]}'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if overlapping or after, False otherwise

        MEOS Functions:
            overright_spanset_span, overright_spanset_spanset,
            overafter_periodset_timestamp, overafter_periodset_timestampset,
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overafter_periodset_timestamp(self._inner,
                datetime_to_timestamptz(other))
        elif isinstance(other, Temporal):
            return self.is_over_or_right(other.period())
        elif isinstance(other, get_args(Box)):
            return self.is_over_or_right(other.to_period())
        else:
            return super().is_over_or_right(other)

    def is_right(self, other: Union[Time, Box, Temporal]) -> bool:
        """
        Returns whether ``self`` is strictly after ``other``.That is, ``self``
        starts after ``other`` ends.

        Examples:
            >>> PeriodSet('{[2012-01-02, 2012-01-03]}').is_right(PeriodSet('{[2012-01-01, 2012-01-02)}'))
            >>> True
            >>> PeriodSet('{(2012-01-02, 2012-01-03]}').is_right(PeriodSet('{[2012-01-01, 2012-01-02)}'))
            >>> True
            >>> PeriodSet('{[2012-01-02, 2012-01-03]}').is_right(PeriodSet('{[2012-01-01, 2012-01-02]}'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if after, False otherwise

        MEOS Functions:
            right_spanset_span, right_spanset_spanset,
            overbefore_timestamp_periodset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overbefore_timestamp_periodset(datetime_to_timestamptz(other),
                self._inner)
        elif isinstance(other, Temporal):
            return self.is_right(other.period())
        elif isinstance(other, get_args(Box)):
            return self.is_right(other.to_period())
        else:
            return super().is_right(other)

    # ------------------------- Distance Operations ---------------------------
    def distance(self, other: Union[Time, Box, Temporal]) -> timedelta:
        """
        Returns the temporal distance between ``self`` and ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            A :class:`datetime.timedelta` instance

        MEOS Functions:
            distance_periodset_period, distance_periodset_periodset, 
            distance_periodset_timestamp, distance_periodset_timestampset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return timedelta(seconds=distance_periodset_timestamp(self._inner,
                datetime_to_timestamptz(other)))
        if isinstance(other, Temporal):
            return self.distance(other.period())
        elif isinstance(other, get_args(Box)):
            return self.distance(other.to_period())
        else:
            return timedelta(seconds=super().distance(other))

    # ------------------------- Set Operations --------------------------------
    @overload
    def intersection(self, other: Period) -> PeriodSet:
        ...

    @overload
    def intersection(self, other: PeriodSet) -> PeriodSet:
        ...

    @overload
    def intersection(self, other: datetime) -> datetime:
        ...

    def intersection(self, other: Time) -> Union[PeriodSet, datetime, TimestampSet]:
        """
        Returns the temporal intersection of ``self`` and ``other``.

        Args:
            other: temporal object to intersect with

        Returns:
            A :class:`Time` instance. The actual class depends on ``other``.

        MEOS Functions:
            intersection_periodset_timestamp, intersection_spanset_spanset,
            intersection_spanset_span
        """
        if isinstance(other, datetime):
            result = intersection_periodset_timestamp(self._inner,
                datetime_to_timestamptz(other))
            return timestamptz_to_datetime(result) if result is not None else None
        else:
            result = super().intersection(other)
        return PeriodSet(_inner=result) if result is not None else None

    def __mul__(self, other):
        """
        Returns the temporal intersection of ``self`` and ``other``.

        Args:
            other: temporal object to intersect with

        Returns:
            A :class:`Time` instance. The actual class depends on ``other``.

        MEOS Functions:
            intersection_periodset_timestamp, intersection_spanset_spanset,
            intersection_spanset_span
        """
        return self.intersection(other)

    def minus(self, other: Time) -> PeriodSet:
        """
        Returns the temporal difference of ``self`` and ``other``.

        Args:
            other: temporal object to diff with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
            minus_spanset_span, minus_spanset_spanset, minus_periodset_timestamp
        """
        if isinstance(other, datetime):
            result = minus_periodset_timestamp(self._inner,
                datetime_to_timestamptz(other))
        else:
            result = super().minus(other)
        return PeriodSet(_inner=result) if result is not None else None

    def __sub__(self, other):
        """
        Returns the temporal difference of ``self`` and ``other``.

        Args:
            other: temporal object to diff with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
            minus_spanset_span, minus_spanset_spanset,
            minus_periodset_timestamp
        """
        return self.minus(other)

    def union(self, other: Time) -> PeriodSet:
        """
        Returns the temporal union of ``self`` and ``other``.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
            union_periodset_timestamp, union_spanset_spanset,
            union_spanset_span
        """
        if isinstance(other, datetime):
            result = union_periodset_timestamp(self._inner,
                datetime_to_timestamptz(other))
        else:
            result = super().union(other)
        return PeriodSet(_inner=result) if result is not None else None

    def __add__(self, other):
        """
        Returns the temporal union of ``self`` and ``other``.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
            union_periodset_timestamp, union_spanset_spanset,
            union_spanset_span
        """
        return self.union(other)

    # ------------------------- Plot Operations -------------------------------
    def plot(self, *args, **kwargs):
        from ...plotters import TimePlotter
        return TimePlotter.plot_periodset(self, *args, **kwargs)



from typing import Union, List

from .point_sequence_plotter import TemporalPointSequencePlotter
from .. import TPointSeq
from ..main import TPointSeqSet


class TemporalPointSequenceSetPlotter:
    """
    Plotter for :class:`TPointSeqSet` and lists of :class:`TPointSeq`.
    """

    @staticmethod
    def plot_xy(sequence_set: Union[TPointSeqSet, List[TPointSeq]], *args, **kwargs):
        """
        Plot a TPointSeqSet or a list of TPointSeq on the given axes. Every sequence in the set will be plotted with the
        same color.

        Params:
            sequence_set: The :class:`TPointSeqSet` or list of :class:`TPointSeq` to plot.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List of lists with the plotted elements of each sequence.

        See Also:
            :func:`~pymeos.plotters.point_sequence_plotter.TemporalPointSequencePlotter.plot_xy`,
            :func:`~pymeos.plotters.point_sequence_plotter.TemporalPointSequencePlotter.plot_sequences_xy`
        """
        seqs = sequence_set.sequences() if isinstance(sequence_set, TPointSeqSet) else sequence_set
        plots = [TemporalPointSequencePlotter.plot_xy(seqs[0], *args, **kwargs)]
        if 'color' not in kwargs:
            kwargs['color'] = plots[0][0][0].get_color()
        kwargs.pop('label', None)
        for seq in seqs[1:]:
            plots.append(TemporalPointSequencePlotter.plot_xy(seq, *args,  **kwargs))
        return plots



from pymeos_cffi import *

from .aggregator import BaseAggregator
from ..boxes import TBox
from ..main import TInt, TFloat, TNumber


class TemporalAverageAggregator(BaseAggregator[TNumber, TNumber]):
    """
    Temporal average of aggregated temporal numbers. Accepts both Temporal Integers and Temporal floats

    MEOS Functions:
        tnumber_tavg_transfn, tnumber_tavg_finalfn
    """
    _add_function = tnumber_tavg_transfn
    _final_function = tnumber_tavg_finalfn


class TemporalNumberExtentAggregator(BaseAggregator[TNumber, TBox]):
    """
    Temporal and numeric extent of aggregated temporal numbers, i.e. smallest :class:`~pymeos.time.boxes.TBox` that
    includes all aggregated temporal numbers.

    MEOS Functions:
        tnumber_extent_transfn, temporal_tagg_finalfn
    """
    _add_function = tnumber_extent_transfn

    @classmethod
    def _finish(cls, state) -> TBox:
        return TBox(_inner=state)


class TemporalIntMaxAggregator(BaseAggregator[TInt, TInt]):
    """
    Temporal maximum of all aggregated temporal integers.

    MEOS Functions:
        tint_tmax_transfn, temporal_tagg_finalfn
    """
    _add_function = tint_tmax_transfn


class TemporalIntMinAggregator(BaseAggregator[TInt, TInt]):
    """
    Temporal minimum of all aggregated temporal integers.

    MEOS Functions:
        tint_tmin_transfn, temporal_tagg_finalfn
    """
    _add_function = tint_tmin_transfn


class TemporalIntSumAggregator(BaseAggregator[TInt, TInt]):
    """
    Temporal summ of all aggregated temporal integers.

    MEOS Functions:
        tint_tsum_transfn, temporal_tagg_finalfn
    """
    _add_function = tint_tsum_transfn


class TemporalFloatMaxAggregator(BaseAggregator[TFloat, TFloat]):
    """
    Temporal maximum of all aggregated temporal floats.

    MEOS Functions:
        tfloat_tmax_transfn, temporal_tagg_finalfn
    """
    _add_function = tfloat_tmax_transfn


class TemporalFloatMinAggregator(BaseAggregator[TFloat, TFloat]):
    """
    Temporal minimum of all aggregated temporal floats.

    MEOS Functions:
        tfloat_tmin_transfn, temporal_tagg_finalfn
    """
    _add_function = tfloat_tmin_transfn


class TemporalFloatSumAggregator(BaseAggregator[TFloat, TFloat]):
    """
    Temporal summ of all aggregated temporal floats.

    MEOS Functions:
        tfloat_tsum_transfn, temporal_tagg_finalfn
    """
    _add_function = tfloat_tsum_transfn



import asyncpg

from pymeos import TBool, TInt, TFloat, TText, TGeomPoint, TGeogPoint, TBox, STBox, TimestampSet, Period, PeriodSet


class MobilityDB:
    """
    Helper class to register MobilityDB classes to an asyncpg connection and their automatic conversion to
    PyMEOS classes.
    """
    @classmethod
    async def connect(cls, *args, **kwargs) -> asyncpg.connection.Connection:
        """
        A coroutine to establish a connection to a MobilityDB server.

        Refer to :func:`asyncpg.connection.connect` for the list of valid arguments.

        If the connection already exists, see the :func:`~pymeos.db.asyncpg.MobilityDB.register` in this class.

        Args:
            *args: positional arguments that will be passed to :func:`asyncpg.connection.connect`
            **kwargs: keyword arguments that will be passed to :func:`asyncpg.connection.connect`

        Returns:
            A new :class:`asyncpg.connection.Connection` object with the MobilityDB classes registered.

        """
        conn = await asyncpg.connect(*args, **kwargs)
        await cls.register(conn)
        return conn

    @classmethod
    async def register(cls, connection: asyncpg.connection.Connection) -> None:
        """
        Registers MobilityDB classes to the passed connection.

        Args:
            connection: An :class:`asyncpg.connection.Connection` to register the classes to.
        """
        classes = [TimestampSet, Period, PeriodSet, TBox, TBool, TInt, TFloat, TText, STBox, TGeomPoint, TGeogPoint]
        for cl in classes:
            await connection.set_type_codec(cl._mobilitydb_name, encoder=str, decoder=cl.read_from_cursor)



from .box_plotter import BoxPlotter
from .point_sequence_plotter import TemporalPointSequencePlotter
from .point_sequenceset_plotter import TemporalPointSequenceSetPlotter
from .range_plotter import SpanPlotter
from .sequence_plotter import TemporalSequencePlotter
from .sequenceset_plotter import TemporalSequenceSetPlotter
from .time_plotter import TimePlotter



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, List, Union, TYPE_CHECKING, Set, Generic, TypeVar, Type

from pandas import DataFrame
from pymeos_cffi import *

from .interpolation import TInterpolation
from ..collections import *

if TYPE_CHECKING:
    from .tinstant import TInstant
    from .tsequence import TSequence
    from ..main import TBool
    from ..boxes import Box
TBase = TypeVar('TBase')
TG = TypeVar('TG', bound='Temporal[Any]')
TI = TypeVar('TI', bound='TInstant[Any]')
TS = TypeVar('TS', bound='TSequence[Any]')
TSS = TypeVar('TSS', bound='TSequenceSet[Any]')
Self = TypeVar('Self', bound='Temporal[Any]')


class Temporal(Generic[TBase, TG, TI, TS, TSS], ABC):
    """
    Abstract class for representing temporal values of any subtype.
    """
    __slots__ = ['_inner']

    BaseClass = None
    """
    Class of the base type, for example, ``float`` for ``TFloat``
    """

    ComponentClass = None
    """
    Class of the components, for example, 

    1. ``TFloatInst`` for both ``TFloatInst`` and ``TFloatSeq``
    2. ``TFloatSeq`` for ``TFloatSeqSet``.
    """

    _parse_function = None

    # ------------------------- Constructors ----------------------------------
    @classmethod
    def _factory(cls, inner):
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(inner)

    def __copy__(self) -> Self:
        """
        Returns a copy of the temporal object.
        """
        inner_copy = temporal_copy(self._inner)
        return self.__class__._factory(inner_copy)

    # ------------------------- Output ----------------------------------------
    def __str__(self) -> str:
        """
        Returns the string representation of the `self`.
        """
        pass

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}'
                f'({self})')

    @staticmethod
    @abstractmethod
    def from_base_time(value: TBase, base: Time) -> TG:
        """
        Create a temporal object from a boolean value and a time object.

        Args:
            value: Base value.
            base: Time object to use as temporal dimension.

        Returns:
            A new temporal object.

        """
        pass

    @classmethod
    def from_wkb(cls: Type[Self], wkb: bytes) -> Self:
        """
        Returns a temporal object from WKB bytes.

        Args:
            wkb: The WKB string.

        Returns:
            A temporal object from WKB bytes.

        MEOS Functions:
            temporal_from_wkb
        """
        result = temporal_from_wkb(wkb)
        return Temporal._factory(result)

    @classmethod
    def from_hexwkb(cls: Type[Self], hexwkb: str) -> Self:
        """
        Returns a temporal object from a hex-encoded WKB string.

        Args:
            hexwkb: The hex-encoded WKB string.

        Returns:
            A temporal object from a hex-encoded WKB string.

        MEOS Functions:
            temporal_from_hexwkb
        """
        result = temporal_from_hexwkb(hexwkb)
        return Temporal._factory(result)

    @classmethod
    def from_mfjson(cls: Type[Self], mfjson: str) -> Self:
        """
        Returns a temporal object from a MF-JSON string.

        Args:
            mfjson: The MF-JSON string.

        Returns:
            A temporal object from a MF-JSON string.

        MEOS Functions:
            temporal_from_mfjson
        """
        result = temporal_from_mfjson(mfjson)
        return Temporal._factory(result)

    @classmethod
    def from_merge(cls: Type[Self], *temporals: TG) -> Self:
        """
        Returns a temporal object that is the result of merging the given
        temporal objects.

        Args:
            *temporals: The temporal objects to merge.

        Returns:
            A temporal object that is the result of merging the given temporal
            objects.

        MEOS Functions:
            temporal_merge_array
        """
        result = temporal_merge_array([temp._inner for temp in temporals \
                                       if temp is not None], len(temporals))
        return Temporal._factory(result)

    @classmethod
    def from_merge_array(cls: Type[Self], temporals: List[TG]) -> Self:
        """
        Returns a temporal object that is the result of merging the given
        temporal objects.

        Args:
            temporals: The temporal objects to merge.

        Returns:
            A temporal object that is the result of merging the given temporal
            objects.
        """
        result = temporal_merge_array([temp._inner for temp in temporals \
                                       if temp is not None], len(temporals))
        return Temporal._factory(result)

    # ------------------------- Output ----------------------------------------
    @abstractmethod
    def as_wkt(self) -> str:
        """
        Returns the temporal object as a WKT string.

        Returns:
            The temporal object as a WKT string.
        """
        pass

    def as_mfjson(self, with_bbox: bool = True, flags: int = 3,
                  precision: int = 6, srs: Optional[str] = None) -> str:
        """
        Returns the temporal object as a MF-JSON string.

        Args:
            with_bbox: Whether to include the bounding box in the output.
            flags: The flags to use for the output.
            precision: The precision to use for the output.
            srs: The SRS to use for the output.

        Returns:
            The temporal object as a MF-JSON string.

        MEOS Functions:
            temporal_as_mfjson
        """
        return temporal_as_mfjson(self._inner, with_bbox, flags, precision, srs)

    def as_wkb(self) -> bytes:
        """
        Returns the temporal object as a hex-encoded WKB string.

        Returns:
            The temporal object as a hex-encoded WKB string.

        MEOS Functions:
            temporal_as_hexwkb
        """
        return temporal_as_wkb(self._inner, 4)

    def as_hexwkb(self) -> str:
        """
        Returns the temporal object as a hex-encoded WKB string.

        Returns:
            The temporal object as a hex-encoded WKB string.

        MEOS Functions:
            temporal_as_hexwkb
        """
        return temporal_as_hexwkb(self._inner, 4)[0]

    # ------------------------- Accessors -------------------------------------
    def bounding_box(self) -> Union[Period, Box]:
        """
        Returns the bounding box of `self`.

        Returns:
            The bounding box of `self`.

        MEOS Functions:
            temporal_to_period
        """
        return Period(_inner=temporal_to_period(self._inner))

    def interpolation(self) -> TInterpolation:
        """
        Returns the interpolation of `self`

        MEOS Functions:
            temporal_interpolation
        """
        val = temporal_interp(self._inner)
        return TInterpolation.from_string(val)

    @abstractmethod
    def value_set(self) -> Set[TBase]:
        """
        Returns the unique values in `self`.
        """
        pass

    def values(self) -> List[TBase]:
        """
        Returns the list of values taken by `self`.
        """
        return [i.value() for i in self.instants()]

    @abstractmethod
    def start_value(self) -> TBase:
        """
        Returns the starting value of `self`.
        """
        pass

    @abstractmethod
    def end_value(self) -> TBase:
        """
        Returns the end value of `self`.
        """
        pass

    def min_value(self) -> TBase:
        """
        Returns the minimum value that `self` takes.
        """
        return min(self.value_set())

    def max_value(self) -> TBase:
        """
        Returns the maximum value that `self` takes.
        """
        return max(self.value_set())

    @abstractmethod
    def value_at_timestamp(self, timestamp: datetime) -> TBase:
        """
        Returns the value that `self` takes at a certain moment.
        """
        pass

    def time(self) -> PeriodSet:
        """
        Returns the :class:`PeriodSet` on which `self` is defined.

        MEOS Functions:
            temporal_time
        """
        return PeriodSet(_inner=temporal_time(self._inner))

    def duration(self, ignore_gaps=False) -> timedelta:
        """
        Returns the duration of `self`. By default, the gaps in `self` are
        taken into account, but this can be changed by setting `ignore_gaps`
        to ``True``. This will only potentially alter the result for sequence
        sets and discrete sequences.

        Parameters:
            ignore_gaps: Whether to take into account potential time gaps in
            the temporal value.

        MEOS Functions:
            temporal_duration
        """
        return interval_to_timedelta(temporal_duration(self._inner, ignore_gaps))

    def period(self) -> Period:
        """
        Returns the :class:`Period` on which `self` is defined ignoring
        potential time gaps.

        MEOS Functions:
            temporal_to_period
        """
        return self.timespan()

    def timespan(self) -> Period:
        """
        Returns the :class:`Period` on which `self` is defined ignoring
        potential time gaps.

        MEOS Functions:
            temporal_to_period
        """
        return Period(_inner=temporal_to_period(self._inner))

    def num_instants(self) -> int:
        """
        Returns the number of instants in `self`.

        MEOS Functions:
            temporal_num_instants
        """
        return temporal_num_instants(self._inner)

    def start_instant(self) -> TI:
        """
        Returns the first instant in `self`.

        MEOS Functions:
            temporal_start_instant
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(temporal_start_instant(self._inner))

    def end_instant(self) -> TI:
        """
        Returns the last instant in `self`.

        MEOS Functions:
            temporal_end_instant
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(temporal_end_instant(self._inner))

    def min_instant(self) -> TI:
        """
        Returns the instant in `self` with the minimum value.
        If multiple instants have the minimum value, the first one is returned.

        MEOS Functions:
            temporal_min_instant
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(temporal_min_instant(self._inner))

    def max_instant(self) -> TI:
        """
        Returns the instant in `self` with the maximum value.
        If multiple instants have the maximum value, the first one is returned.

        MEOS Functions:
            temporal_max_instant
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(temporal_max_instant(self._inner))

    def instant_n(self, n: int) -> TI:
        """
        Returns the n-th instant in `self`. (0-based)

        MEOS Functions:
            temporal_instant_n
        """
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(temporal_instant_n(self._inner, n + 1))

    def instants(self) -> List[TI]:
        """
        Returns the instants in `self`.

        MEOS Functions:
            temporal_instants
        """
        from ..factory import _TemporalFactory
        ins, count = temporal_instants(self._inner)
        return [_TemporalFactory.create_temporal(ins[i]) for i in range(count)]

    def num_timestamps(self) -> int:
        """
        Returns the number of timestamps in `self`.

        MEOS Functions:
            temporal_num_timestamps
        """
        return temporal_num_timestamps(self._inner)

    def start_timestamp(self) -> datetime:
        """
        Returns the first timestamp in `self`.

        MEOS Functions:
            temporal_start_timestamps
        """
        return timestamptz_to_datetime(temporal_start_timestamp(self._inner))

    def end_timestamp(self) -> datetime:
        """
        Returns the last timestamp in `self`.

        MEOS Functions:
            temporal_end_timestamps
        """
        return timestamptz_to_datetime(temporal_end_timestamp(self._inner))

    def timestamp_n(self, n: int) -> datetime:
        """
        Returns the n-th timestamp in `self`. (0-based)

        MEOS Functions:
            temporal_timestamp_n
        """
        return timestamptz_to_datetime(temporal_timestamp_n(self._inner, n + 1))

    def timestamps(self) -> List[datetime]:
        """
        Returns the timestamps in `self`.

        MEOS Functions:
            temporal_timestamps
        """
        ts, count = temporal_timestamps(self._inner)
        return [timestamptz_to_datetime(ts[i]) for i in range(count)]

    def segments(self) -> List[TS]:
        """
        Returns the temporal segments in `self`.

        MEOS Functions:
            temporal_segments
        """
        seqs, count = temporal_segments(self._inner)
        from ..factory import _TemporalFactory
        return [_TemporalFactory.create_temporal(seqs[i]) for i in range(count)]

    def __hash__(self) -> int:
        """
        Returns the hash of the temporal object.

        Returns:
            The hash of the temporal object.

        MEOS Functions:
            temporal_hash
        """
        return temporal_hash(self._inner)

    # ------------------------- Transformations -------------------------------
    def set_interpolation(self: Self, interpolation: TInterpolation) -> Self:
        """
        Returns a new :class:`Temporal` object equal to `self` with the given
        interpolation.

        MEOS Functions:
            temporal_set_interpolation
        """
        new_temp = temporal_set_interp(self._inner, interpolation)
        return Temporal._factory(new_temp)

    def shift_time(self, delta: timedelta) -> Self:
        """
        Returns a new :class:`Temporal` with the temporal dimension shifted by
        ``delta``.

        Args:
            delta: :class:`datetime.timedelta` instance to shift

        MEOS Functions:
            temporal_shift_time
        """
        shifted = temporal_shift_time(self._inner, timedelta_to_interval(delta))
        return Temporal._factory(shifted)

    def scale_time(self, duration: timedelta) -> Self:
        """
        Returns a new :class:`Temporal` scaled so the temporal dimension has
        duration ``duration``.

        Args:
            duration: :class:`datetime.timedelta` instance representing the
            duration of the new temporal

        MEOS Functions:
            temporal_scale_time
        """
        scaled = temporal_scale_time(self._inner, timedelta_to_interval(duration))
        return Temporal._factory(scaled)

    def shift_scale_time(self, shift: Optional[timedelta] = None,
                         duration: Optional[timedelta] = None) -> Self:
        """
        Returns a new :class:`Temporal` with the time dimension shifted by
        ``shift`` and scaled so the temporal dimension has duration
        ``duration``.

        Args:
            shift: :class:`datetime.timedelta` instance to shift
            duration: :class:`datetime.timedelta` instance representing the
            duration of the new temporal

        MEOS Functions:
            temporal_shift_scale_time
        """
        assert shift is not None or duration is not None, \
            'shift and duration must not be both None'
        scaled = temporal_shift_scale_time(
            self._inner,
            timedelta_to_interval(shift) if shift else None,
            timedelta_to_interval(duration) if duration else None
        )
        return Temporal._factory(scaled)

    def temporal_sample(self, duration: Union[str, timedelta],
                        start: Optional[Union[str, datetime]] = None) -> TG:
        """
        Returns a new :class:`Temporal` downsampled with respect to ``duration``.

        Args:
            duration: A :class:`str` or :class:`timedelta` with the duration of
                the temporal tiles.
            start: A :class:`str` or :class:`datetime` with the start time of
                the temporal tiles. If None, the start time used by default is
                Monday, January 3, 2000.
        MEOS Functions:
            temporal_tsample
        """
        if start is None:
            st = pg_timestamptz_in('2000-01-03', -1)
        elif isinstance(start, datetime):
            st = datetime_to_timestamptz(start)
        else:
            st = pg_timestamptz_in(start, -1)
        if isinstance(duration, timedelta):
            dt = timedelta_to_interval(duration)
        else:
            dt = pg_interval_in(duration, -1)
        result = temporal_tsample(self._inner, dt, st)
        return Temporal._factory(result)

    def temporal_precision(self, duration: Union[str, timedelta],
                           start: Optional[Union[str, datetime]] = None) -> TG:
        """
        Returns a new :class:`Temporal` with precision reduced to ``duration``.

        Args:
            duration: A :class:`str` or :class:`timedelta` with the duration
                of the temporal tiles.
            start: A :class:`str` or :class:`datetime` with the start time of
                the temporal tiles. If None, the start time used by default is
                Monday, January 3, 2000.

        MEOS Functions:
            temporal_tprecision
        """
        if start is None:
            st = pg_timestamptz_in('2000-01-03', -1)
        elif isinstance(start, datetime):
            st = datetime_to_timestamptz(start)
        else:
            st = pg_timestamptz_in(start, -1)
        if isinstance(duration, timedelta):
            dt = timedelta_to_interval(duration)
        else:
            dt = pg_interval_in(duration, -1)
        result = temporal_tprecision(self._inner, dt, st)
        return Temporal._factory(result)

    def to_instant(self) -> TI:
        """
        Returns `self` as a :class:`TInstant`.

        MEOS Functions:
            temporal_to_tinstant
        """
        inst = temporal_to_tinstant(self._inner)
        return Temporal._factory(inst)

    def to_sequence(self, interpolation: TInterpolation) -> TS:
        """
        Converts `self` into a :class:`TSequence`.

        MEOS Functions:
            temporal_to_sequence
        """
        seq = temporal_to_tsequence(self._inner, interpolation)
        return Temporal._factory(seq)

    def to_sequenceset(self, interpolation: TInterpolation) -> TSS:
        """
        Returns `self` as a :class:`TSequenceSet`.

        MEOS Functions:
            temporal_to_tsequenceset
        """
        ss = temporal_to_tsequenceset(self._inner, interpolation)
        return Temporal._factory(ss)

    def to_dataframe(self) -> DataFrame:
        """
        Returns `self` as a :class:`pd.DataFrame` with two columns: `time`
            and `value`.
        """
        data = {
            'time': self.timestamps(),
            'value': self.values()
        }
        return DataFrame(data).set_index(keys='time')

    # ------------------------- Modifications ---------------------------------
    def append_instant(self, instant: TInstant[TBase], max_dist: Optional[float] = 0.0,
                       max_time: Optional[timedelta] = None) -> TG:
        """
        Returns a new :class:`Temporal` object equal to `self` with `instant`
        appended.

        Args:
            instant: :class:`TInstant` to append
            max_dist: Maximum distance for defining a gap
            max_time: Maximum time for defining a gap

        MEOS Functions:
            temporal_append_tinstant
        """
        if max_time is None:
            interv = None
        else:
            interv = timedelta_to_interval(max_time)
        new_inner = temporal_append_tinstant(self._inner, instant._inner,
                                             max_dist, interv, False)
        return Temporal._factory(new_inner)

    def append_sequence(self, sequence: TSequence[TBase]) -> TG:
        """
        Returns a new :class:`Temporal` object equal to `self` with `sequence`
        appended.

        Args:
            sequence: :class:`TSequence` to append

        MEOS Functions:
            temporal_append_tsequence
        """
        new_inner = temporal_append_tsequence(self._inner, sequence._inner, False)
        return Temporal._factory(new_inner)

    def merge(self, other: Union[type(None), Temporal[TBase], List[Temporal[TBase]]]) -> TG:
        """
        Returns a new :class:`Temporal` object that is the result of merging
        `self` with `other`.

        MEOS Functions:
            temporal_merge, temporal_merge_array
        """
        if isinstance(self, type(None)):
            return other
        if isinstance(other, type(None)):
            return self
        if isinstance(other, Temporal):
            new_temp = temporal_merge(self._inner, other._inner)
        elif isinstance(other, list):
            new_temp = temporal_merge_array([self._inner,
                                             *(o._inner for o in other)], len(other) + 1)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(new_temp)

    def insert(self, other: TG, connect: bool = True) -> TG:
        """
        Returns a new :class:`Temporal` object equal to `self` with `other`
        inserted.

        Args:
            other: :class:`Temporal` object to insert in `self`
            connect: wether to connect the inserted elements with the existing
                elements.

        MEOS Functions:
            temporal_insert
        """
        new_inner = temporal_insert(self._inner, other._inner, connect)
        if new_inner == self._inner:
            return self
        return Temporal._factory(new_inner)

    def update(self, other: TG, connect: bool = True) -> TG:
        """
        Returns a new :class:`Temporal` object equal to `self` updated with
        `other`.

        Args:
            other: :class:`Temporal` object to update `self` with
            connect: wether to connect the updated elements with the
                existing elements.

        MEOS Functions:
            temporal_update
        """
        new_inner = temporal_update(self._inner, other._inner, connect)
        if new_inner == self._inner:
            return self
        return Temporal._factory(new_inner)

    def delete(self, other: Time, connect: bool = True) -> TG:
        """
        Returns a new :class:`Temporal` object equal to `self` with elements at
        `other` removed.

        Args:
            other: :class:`Time` object to remove from `self`
            connect: whether to connect the potential gaps generated by the
                deletions.

        MEOS Functions:
            temporal_update
        """
        if isinstance(other, datetime):
            new_inner = temporal_delete_timestamp(self._inner,
                                                  datetime_to_timestamptz(other), connect)
        elif isinstance(other, TimestampSet):
            new_inner = temporal_delete_timestampset(self._inner, other._inner,
                                                     connect)
        elif isinstance(other, Period):
            new_inner = temporal_delete_period(self._inner, other._inner,
                                               connect)
        elif isinstance(other, PeriodSet):
            new_inner = temporal_delete_periodset(self._inner, other._inner,
                                                  connect)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        if new_inner == self._inner:
            return self
        return Temporal._factory(new_inner)

    # ------------------------- Restrictions ----------------------------------
    def at(self, other: Time) -> TG:
        """
        Returns a new temporal object with the values of `self` restricted to
        the time `other`.

        Args:
            other: A time object to restrict the values of `self` to.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            temporal_at_timestamp, temporal_at_timestampset,
            temporal_at_period, temporal_at_periodset
        """
        if isinstance(other, datetime):
            result = temporal_at_timestamp(self._inner,
                                           datetime_to_timestamptz(other))
        elif isinstance(other, TimestampSet):
            result = temporal_at_timestampset(self._inner, other._inner)
        elif isinstance(other, Period):
            result = temporal_at_period(self._inner, other._inner)
        elif isinstance(other, PeriodSet):
            result = temporal_at_periodset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def at_min(self) -> TG:
        """
        Returns a new temporal object containing the times ``self`` is at its
        minimum value.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            temporal_at_min
        """
        result = temporal_at_min(self._inner)
        return Temporal._factory(result)

    def at_max(self) -> TG:
        """
        Returns a new temporal object containing the times ``self`` is at its
        maximum value.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            temporal_at_max
        """
        result = temporal_at_max(self._inner)
        return Temporal._factory(result)

    def minus(self, other: Time) -> TG:
        """
        Returns a new temporal object with the values of `self` removing those
        happening at `other`.

        Args:
            other: A time object to remove from `self`.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            temporal_minus_timestamp, temporal_minus_timestampset,
            temporal_minus_period, temporal_minus_periodset
        """
        if isinstance(other, Period):
            result = temporal_minus_period(self._inner, other._inner)
        elif isinstance(other, PeriodSet):
            result = temporal_minus_periodset(self._inner, other._inner)
        elif isinstance(other, datetime):
            result = temporal_minus_timestamp(self._inner,
                                              datetime_to_timestamptz(other))
        elif isinstance(other, TimestampSet):
            result = temporal_minus_timestampset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def minus_min(self) -> TG:
        """
        Returns a new temporal object containing the times ``self`` is not at
        its minimum value.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            temporal_minus_min
        """
        result = temporal_minus_min(self._inner)
        return Temporal._factory(result)

    def minus_max(self) -> TG:
        """
        Returns a new temporal object containing the times ``self`` is not at
        its maximum value.

        Returns:
            A new temporal object of the same subtype as `self`.

        MEOS Functions:
            temporal_minus_max
        """
        result = temporal_minus_max(self._inner)
        return Temporal._factory(result)

    # ------------------------- Topological Operations ------------------------
    def is_adjacent(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is adjacent to the bounding
        box of `other`. Temporal subclasses may override this method to provide
        more specific behavior related to their types an check adjacency over
        more dimensions.

        Args:
            other: A time or temporal object to compare to `self`.

        Returns:
            True if adjacent, False otherwise.

        See Also:
            :meth:`Period.is_adjacent`
        """
        return self.bounding_box().is_adjacent(other)

    def is_temporally_adjacent(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether the bounding period of `self` is temporally adjacent
        to the bounding period of `other`.

        Args:
            other: A time or temporal object to compare to `self`.

        Returns:
            True if adjacent, False otherwise.

        See Also:
            :meth:`Period.is_adjacent`
        """
        return self.period().is_adjacent(other)

    def is_contained_in(self, container: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether the bounding period of `self` is contained in the
        bounding period of `container`. Temporal subclasses may override this
        method to provide more specific behavior related to their types

        Args:
            container: A time or temporal object to compare to `self`.

        Returns:
            True if contained, False otherwise.

        See Also:
            :meth:`Period.is_contained_in`
        """
        return self.bounding_box().is_contained_in(container)

    def is_temporally_contained_in(self,
                                   container: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether the bounding period of `self` is contained in the
        bounding period of `container`.

        Args:
            container: A time or temporal object to compare to `self`.

        Returns:
            True if contained, False otherwise.

        See Also:
            :meth:`Period.is_contained_in`
        """
        return self.period().is_contained_in(container)

    def contains(self, content: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether the bounding period of `self` contains the bounding
        period of `content`. Temporal subclasses may override this method to
        provide more specific behavior related to their types

        Args:
            content: A time or temporal object to compare to `self`.

        Returns:
            True if contains, False otherwise.

        See Also:
            :meth:`Period.contains`
        """
        return self.bounding_box().contains(content)

    def __contains__(self, item):
        """
        Returns whether the bounding period of `self` contains the bounding
        period of `content`.

        Args:
            item: A time or temporal object to compare to `self`.

        Returns:
            True if contains, False otherwise.

        See Also:
            :meth:`Period.contains`
        """
        return self.contains(item)

    def temporally_contains(self, content: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether the bounding period of `self` contains the bounding
        period of `content`.

        Args:
            content: A time or temporal object to compare to `self`.

        Returns:
            True if contains, False otherwise.

        See Also:
            :meth:`Period.contains`
        """
        return self.period().contains(content)

    def overlaps(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether the bounding period of `self` overlaps the bounding
        period of `other`. Temporal subclasses may override this method to
        provide more specific behavior related to their types

        Args:
            other: A time or temporal object to compare to `self`.

        Returns:
            True if overlaps, False otherwise.

        See Also:
            :meth:`Period.overlaps`
        """
        return self.bounding_box().overlaps(other)

    def temporally_overlaps(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether the bounding period of `self` overlaps the bounding
        period of `other`.

        Args:
            other: A time or temporal object to compare to `self`.

        Returns:
            True if overlaps, False otherwise.

        See Also:
            :meth:`Period.overlaps`
        """
        return self.period().overlaps(other)

    def is_same(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether the bounding period of `self` is the same as the
        bounding period of `other`. Temporal subclasses may override this
        method to provide more specific behavior related to their types

        Args:
            other: A time or temporal object to compare to `self`.

        Returns:
            True if same, False otherwise.

        See Also:
            :meth:`Period.is_same`
        """
        return self.bounding_box().is_same(other)

    # ------------------------- Position Operations ---------------------------
    def is_before(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether `self` is before `other`.

        Args:
            other: A time or temporal object to compare `self` to.

        Returns:
            True if `self` is before `other`, False otherwise.

        See Also:
            :meth:`Period.is_before`
        """
        return self.period().is_before(other)

    def is_over_or_before(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether `self` is before `other` allowing overlap. That is,
        `self` doesn't extend after `other`.

        Args:
            other: A time or temporal object to compare `self` to.

        Returns:
            True if `self` is before `other` allowing overlap, False otherwise.

        See Also:
            :meth:`Period.is_over_or_before`
        """
        return self.period().is_over_or_before(other)

    def is_after(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether `self` is after `other`.

        Args:
            other: A time or temporal object to compare `self` to.

        Returns:
            True if `self` is after `other`, False otherwise.

        See Also:
            :meth:`Period.is_after`
        """
        return self.period().is_after(other)

    def is_over_or_after(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether `self` is after `other` allowing overlap. That is,
        `self` doesn't extend before `other`.

        Args:
            other: A time or temporal object to compare `self` to.

        Returns:
            True if `self` is after `other` allowing overlap, False otherwise.

        See Also:
            :meth:`Period.is_over_or_after`
        """
        return self.period().is_over_or_after(other)

    # ------------------------- Similarity Operations -------------------------
    def frechet_distance(self, other: Temporal) -> float:
        """
        Returns the Frechet distance between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`float` with the Frechet distance.

        MEOS Functions:
            temporal_frechet_distance
        """
        return temporal_frechet_distance(self._inner, other._inner)

    def dyntimewarp_distance(self, other: Temporal) -> float:
        """
        Returns the Dynamic Time Warp distance between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`float` with the Dynamic Time Warp distance.

        MEOS Functions:
            temporal_dyntimewarp_distance
        """
        return temporal_dyntimewarp_distance(self._inner, other._inner)

    def hausdorff_distance(self, other: Temporal) -> float:
        """
        Returns the Hausdorff distance between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`float` with the Hausdorff distance.

        MEOS Functions:
            temporal_hausdorff_distance
        """
        return temporal_hausdorff_distance(self._inner, other._inner)

    # ------------------------- Split Operations ------------------------------
    def time_split(self, duration: Union[str, timedelta],
                   start: Optional[Union[str, datetime]] = None) -> List[TG]:
        """
        Returns a list of temporal objects of the same subtype as `self` with
        the same values as `self` but split in temporal tiles of duration
        `duration` starting at `start`.

        Args:
            duration: A :class:`str` or :class:`timedelta` with the duration
                of the temporal tiles.
            start: A :class:`str` or :class:`datetime` with the start time of
                the temporal tiles. If None, the start time used by default is
                Monday, January 3, 2000.

        Returns:
            A list of temporal objects of the same subtype as `self`.

        MEOS Functions:
            temporal_time_split
        """
        if start is None:
            st = pg_timestamptz_in('2000-01-03', -1)
        else:
            st = datetime_to_timestamptz(start) \
                if isinstance(start, datetime) \
                else pg_timestamptz_in(start, -1)
        dt = timedelta_to_interval(duration) \
            if isinstance(duration, timedelta) \
            else pg_interval_in(duration, -1)
        fragments, times, count = temporal_time_split(self._inner, dt, st)
        from ..factory import _TemporalFactory
        return [_TemporalFactory.create_temporal(fragments[i]) \
                for i in range(count)]

    def time_split_n(self, n: int) -> List[TG]:
        """
        Returns a list of temporal objects of the same subtype as `self` with
        the same values as `self` but split in n temporal tiles of equal
        duration.

        Args:
            n: An :class:`int` with the number of temporal tiles.

        Returns:
            A list of temporal objects of the same subtype as `self`.

        MEOS Functions:
            temporal_time_split
        """
        if self.end_timestamp() == self.start_timestamp():
            return [self]
        st = temporal_start_timestamp(self._inner)
        dt = timedelta_to_interval((self.end_timestamp() -
                                    self.start_timestamp()) / n)
        fragments, times, count = temporal_time_split(self._inner, dt, st)
        from ..factory import _TemporalFactory
        return [_TemporalFactory.create_temporal(fragments[i]) \
                for i in range(count)]

    def stops(self, max_distance: Optional[float] = 0.0,
              min_duration: Optional[timedelta] = timedelta()) -> TSS:
        """
        Return the subsequences where the objects stay within an area with a
        given maximum size for at least the specified duration.

        Args:
            max_distance: A :class:`float` with the maximum distance of a stop.
            min_duration: A :class:`timedelta` with the minimum duration of
                a stop.

        Returns:
            A :class:`SequenceSet` of the same subtype as `self` with the stops.

        MEOS Functions:
            temporal_stops
        """
        new_inner = temporal_stops(self._inner, max_distance,
                                   timedelta_to_interval(min_duration))
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(new_inner)

    # ------------------------- Temporal Comparisons --------------------------
    def __comparable(self, other: Temporal) -> bool:
        if not isinstance(other, Temporal):
            return False
        if self.BaseClass == other.BaseClass:
            return True
        if self.BaseClass in [int, float] and other.BaseClass in [int, float]:
            return True
        return False

    def __assert_comparable(self, other: Temporal) -> None:
        if not self.__comparable(other):
            raise TypeError(f'Operation not supported with type {other.__class__}. '
                            f'{self.BaseClass} and {other.BaseClass} are not comparable.')

    def temporal_equal(self, other: Temporal) -> TBool:
        """
        Returns the temporal equality relation between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal equality relation.

        MEOS Functions:
            teq_temporal_temporal
        """
        self.__assert_comparable(other)
        result = teq_temporal_temporal(self._inner, other._inner)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Temporal) -> TBool:
        """
        Returns the temporal not equal relation between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal not equal relation.

        MEOS Functions:
            tne_temporal_temporal
        """
        self.__assert_comparable(other)
        result = tne_temporal_temporal(self._inner, other._inner)
        return Temporal._factory(result)

    def temporal_less(self, other: Temporal) -> TBool:
        """
        Returns the temporal less than relation between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal less than relation.

        MEOS Functions:
            tlt_temporal_temporal
        """
        self.__assert_comparable(other)
        result = tlt_temporal_temporal(self._inner, other._inner)
        return Temporal._factory(result)

    def temporal_less_or_equal(self, other: Temporal) -> TBool:
        """
        Returns the temporal less or equal relation between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal less or equal
            relation.

        MEOS Functions:
            tle_temporal_temporal
        """
        self.__assert_comparable(other)
        result = tle_temporal_temporal(self._inner, other._inner)
        return Temporal._factory(result)

    def temporal_greater_or_equal(self, other: Temporal) -> TBool:
        """
        Returns the temporal greater or equal relation between `self` and
        `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal greater or equal
            relation.

        MEOS Functions:
            tge_temporal_temporal
        """
        self.__assert_comparable(other)
        result = tge_temporal_temporal(self._inner, other._inner)
        return Temporal._factory(result)

    def temporal_greater(self, other: Temporal) -> TBool:
        """
        Returns the temporal greater than relation between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal greater than
            relation.

        MEOS Functions:
            tgt_temporal_temporal
        """
        self.__assert_comparable(other)
        result = tgt_temporal_temporal(self._inner, other._inner)
        return Temporal._factory(result)

    # ------------------------- Comparisons -----------------------------------
    def __eq__(self, other):
        """
        Returns whether `self` is equal to `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`bool` with the result of the equality relation.

        MEOS Functions:
            temporal_eq
        """
        return temporal_eq(self._inner, other._inner)

    def __ne__(self, other):
        """
        Returns whether `self` is not equal to `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`bool` with the result of the not equal relation.

        MEOS Functions:
            temporal_ne
        """
        return temporal_ne(self._inner, other._inner)

    def __lt__(self, other):
        """
        Returns whether `self` is less than `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`bool` with the result of the less than relation.

        MEOS Functions:
            temporal_lt
        """
        return temporal_lt(self._inner, other._inner)

    def __le__(self, other):
        """
        Returns whether `self` is less or equal than `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`bool` with the result of the less or equal than relation.

        MEOS Functions:
            temporal_le
        """
        return temporal_le(self._inner, other._inner)

    def __gt__(self, other):
        """
        Returns whether `self` is greater than `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`bool` with the result of the greater than relation.

        MEOS Functions:
            temporal_gt
        """
        return temporal_gt(self._inner, other._inner)

    def __ge__(self, other):
        """
        Returns whether `self` is greater or equal than `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`bool` with the result of the greater or equal than
            relation.

        MEOS Functions:
            temporal_ge
        """
        return temporal_ge(self._inner, other._inner)



from matplotlib import pyplot as plt

from .range_plotter import SpanPlotter
from .time_plotter import TimePlotter
from ..boxes import TBox, STBox


class BoxPlotter:
    """

    """

    @staticmethod
    def plot_tbox(tbox: TBox, *args, axes=None, **kwargs):
        """
        Plot a TBox on the given axes. If the TBox has only a temporal or spatial dimension, this is equivalent
        to plotting the corresponding Period or Span respectively.

        Params:
            tbox: The :class:`TBox` to plot.
            axes: The axes to plot on. If None, the current axes are used.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.

        See Also:
            :func:`~pymeos.plotters.range_plotter.RangePlotter.plot_range`,
            :func:`~pymeos.plotters.time_plotter.TimePlotter.plot_period`
        """
        if not tbox.has_t:
            return SpanPlotter.plot_span(tbox.to_floatspan(), *args, axes=axes, **kwargs)
        if not tbox.has_x:
            return TimePlotter.plot_period(tbox.to_period(), *args, axes=axes, **kwargs)
        return BoxPlotter._plot_box(tbox.tmin(), tbox.tmax(), tbox.xmin(), tbox.xmax(), *args, axes=axes, **kwargs)

    @staticmethod
    def plot_stbox_xy(stbox: STBox, *args, axes=None, **kwargs):
        """
        Plot an STBox on the given axes. Plots the x and y dimensions.

        Params:
            stbox: The :class:`STBox` to plot.
            axes: The axes to plot on. If None, the current axes are used.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.
        """
        return BoxPlotter._plot_box(stbox.xmin(), stbox.xmax(), stbox.ymin(), stbox.ymax(), *args, axes=axes, **kwargs)

    @staticmethod
    def plot_stbox_xt(stbox: STBox, *args, axes=None, **kwargs):
        """
        Plot an STBox on the given axes. Plots the x and t dimensions.

        Params:
            stbox: The :class:`STBox` to plot.
            axes: The axes to plot on. If None, the current axes are used.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.
        """
        return BoxPlotter._plot_box(stbox.tmin(), stbox.tmax(), stbox.xmin(), stbox.xmax(), *args, axes=axes, **kwargs)

    @staticmethod
    def plot_stbox_yt(stbox: STBox, *args, axes=None, **kwargs):
        """
        Plot an STBox on the given axes. Plots the y and t dimensions.

        Params:
            stbox: The :class:`STBox` to plot.
            axes: The axes to plot on. If None, the current axes are used.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.
        """
        return BoxPlotter._plot_box(stbox.tmin(), stbox.tmax(), stbox.ymin(), stbox.ymax(), *args, axes=axes, **kwargs)

    @staticmethod
    def _plot_box(xmin, xmax, ymin, ymax, *args, axes=None, rotate_xticks=True, draw_filling=True, **kwargs):
        """
        Plot a box on the given axes.

        Params:
            xmin: The minimum x value.
            xmax: The maximum x value.
            ymin: The minimum y value.
            ymax: The maximum y value.
            axes: The axes to plot on. If None, the current axes are used.
            rotate_xticks: Whether to rotate the xticks by 45 degrees.
            draw_filling: Whether to draw a filling.
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.
        """
        base = axes or plt.gca()
        plot = base.plot([xmin, xmax, xmax, xmin, xmin],
                         [ymin, ymin, ymax, ymax, ymin],
                         *args, **kwargs)

        if 'color' not in kwargs:
            kwargs['color'] = plot[0].get_color()
        kwargs.pop('label', None)

        return_array = [plot]

        if draw_filling:
            f = base.fill_between([xmin, xmax], [ymax, ymax], [ymin, ymin], *args, alpha=0.3, **kwargs)
            return_array.append(f)

        if rotate_xticks:
            base.tick_params(axis="x", rotation=45)

        return return_array



from typing import Any

import psycopg
from psycopg import connect
from psycopg.adapt import Loader, Buffer, Dumper

from pymeos import TBool, TInt, TFloat, TText, TGeomPoint, TGeogPoint, TBox, STBox, TimestampSet, Period, PeriodSet


def _pymeos_loader_factory(cl):
    class _PymeosLoader(Loader):
        def load(self, data: Buffer) -> Any:
            return cl.read_from_cursor(data.decode())

    return _PymeosLoader


class _PymeosDumper(Dumper):
    def dump(self, obj: Any) -> Buffer:
        return str(obj).encode()


class MobilityDB:
    """
    Helper class to register MobilityDB classes to a psycopg (3) connection and their automatic conversion to
    PyMEOS classes.
    """
    @classmethod
    def connect(cls, *args, **kwargs):
        """
        Establisesh a connection to a MobilityDB server.

        Refer to :func:`psycopg.connect` for the list of valid arguments.

        If the connection already exists, see the :func:`~pymeos.db.psycopg.MobilityDB.register` in this class.

        Args:
            *args: positional arguments that will be passed to :func:`psycopg.connect`
            **kwargs: keyword arguments that will be passed to :func:`psycopg.connect`

        Returns:
            A new :class:`psycopg.Connection` object with the MobilityDB classes registered.

        """
        connection = connect(*args, **kwargs)
        cls.register(connection)
        return connection

    @classmethod
    def register(cls, connection: psycopg.Connection):
        """
        Registers MobilityDB classes to the passed connection.

        Args:
            connection: An :class:`psycopg.Connection` to register the classes to.
        """
        cursor = connection.cursor()
        classes = [TimestampSet, Period, PeriodSet, TBox, TBool, TInt, TFloat, TText, STBox, TGeomPoint, TGeogPoint]
        for cl in classes:
            cursor.execute(f'SELECT NULL::{cl._mobilitydb_name}')
            oid = cursor.description[0][1]
            connection.adapters.register_loader(oid, _pymeos_loader_factory(cl))
            connection.adapters.register_dumper(cl, _PymeosDumper)



from __future__ import annotations

from abc import ABC
from typing import List, overload, Optional, Union

import shapely as shp
from pymeos_cffi import geoset_start_value, gserialized_to_shapely_geometry, geoset_end_value, geoset_value_n, \
    geoset_values, intersection_geoset_geo, minus_geoset_geo, union_geoset_geo, geoset_as_ewkt, geoset_as_text, \
    geoset_out, geoset_make, geoset_srid, geoset_round, minus_geo_geoset, geomset_in, geogset_in, pgis_geometry_in, \
    geometry_to_gserialized, pgis_geography_in, geography_to_gserialized, intersection_set_set, minus_set_set, \
    union_set_set

from ..base import Set


class GeoSet(Set[shp.Geometry], ABC):
    __slots__ = ['_inner']

    _make_function = geoset_make

    # ------------------------- Constructors ----------------------------------

    # ------------------------- Output ----------------------------------------

    def __str__(self, max_decimals: int = 15):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            geoset_out
        """
        return geoset_out(self._inner, max_decimals)

    def as_ewkt(self, max_decimals: int = 15) -> str:
        """
        Returns the EWKT representation of ``self``.

        Args:
            max_decimals: The number of decimal places to use for the coordinates.

        Returns:
            A :class:`str` instance.

        MEOS Functions:
            geoset_as_ewkt
        """
        return geoset_as_ewkt(self._inner, max_decimals)

    def as_wkt(self, max_decimals: int = 15):
        """
        Returns the WKT representation of ``self``.

        Args:
            max_decimals: The number of decimal places to use for the coordinates.

        Returns:
            A :class:`str` instance.

        MEOS Functions:
            geoset_as_text
        """
        return geoset_as_text(self._inner, max_decimals)

    def as_text(self, max_decimals: int = 15):
        """
        Returns the WKT representation of ``self``.

        Args:
            max_decimals: The number of decimal places to use for the coordinates.

        Returns:
            A :class:`str` instance.

        MEOS Functions:
            geoset_as_text
        """
        return geoset_as_text(self._inner, max_decimals)

    # ------------------------- Conversions -----------------------------------

    def to_spanset(self):
        raise NotImplementedError()

    def to_span(self):
        raise NotImplementedError()

    # ------------------------- Accessors -------------------------------------

    def start_element(self) -> shp.Geometry:
        """
        Returns the first element in ``self``.

        Returns:
            A :class:`Geometry` instance

        MEOS Functions:
            geoset_start_value
        """
        return gserialized_to_shapely_geometry(geoset_start_value(self._inner))

    def end_element(self) -> shp.Geometry:
        """
        Returns the last element in ``self``.

        Returns:
            A :class:`Geometry` instance

        MEOS Functions:
            geoset_end_value
        """
        return gserialized_to_shapely_geometry(geoset_end_value(self._inner))

    def element_n(self, n: int) -> shp.Geometry:
        """
        Returns the ``n``-th element in ``self``.

        Args:
            n: The 0-based index of the element to return.

        Returns:
            A :class:`Geometry` instance

        MEOS Functions:
            geoset_value_n
        """
        super().element_n(n)
        return gserialized_to_shapely_geometry(geoset_value_n(self._inner, n + 1)[0])

    def elements(self) -> List[shp.Geometry]:
        """
        Returns a list of all elements in ``self``.

        Returns:
            A list of :class:`Geometry` instances

        MEOS Functions:
            geoset_values
        """
        elems = geoset_values(self._inner)
        return [gserialized_to_shapely_geometry(elems[i]) for i in range(self.num_elements())]

    def srid(self) -> int:
        """
        Returns the SRID of ``self``.

        Returns:
            An integer

        MEOS Functions:
            geoset_srid
        """
        return geoset_srid(self._inner)

    # ------------------------- Topological Operations --------------------------------

    def contains(self, content: Union[GeoSet, str]) -> bool:
        """
        Returns whether ``self`` contains ``content``.

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise
        """
        return super().contains(content)

    # ------------------------- Set Operations --------------------------------

    @overload
    def intersection(self, other: str) -> Optional[str]:
        ...

    @overload
    def intersection(self, other: GeoSet) -> Optional[GeoSet]:
        ...

    def intersection(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: A :class:`GeoSet` or :class:`Geometry` instance

        Returns:
            An object of the same type as ``other`` or ``None`` if the intersection is empty.

        MEOS Functions:
            intersection_geoset_geo, intersection_set_set
        """
        if isinstance(other, shp.Geometry):
            return gserialized_to_shapely_geometry(
                intersection_geoset_geo(self._inner, geometry_to_gserialized(other))[0])
        elif isinstance(other, GeoSet):
            result = intersection_set_set(self._inner, other._inner)
            return GeoSet(_inner=result) if result is not None else None
        else:
            return super().intersection(other)

    def minus(self, other: Union[GeoSet, shp.Geometry]) -> Optional[GeoSet]:
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: A :class:`GeoSet` or :class:`Geometry` instance

        Returns:
            A :class:`GeoSet` instance or ``None`` if the difference is empty.

        MEOS Functions:
            minus_geoset_geo, minus_set_set

        See Also:
            :meth:`subtract_from`
        """
        if isinstance(other, shp.Geometry):
            result = minus_geoset_geo(self._inner, geometry_to_gserialized(other))
            return GeoSet(_inner=result) if result is not None else None
        elif isinstance(other, GeoSet):
            result = minus_set_set(self._inner, other._inner)
            return GeoSet(_inner=result) if result is not None else None
        else:
            return super().minus(other)

    def subtract_from(self, other: shp.Geometry) -> Optional[shp.Geometry]:
        """
        Returns the difference of ``other`` and ``self``.

        Args:
            other: A :class:`Geometry` instance

        Returns:
            A :class:`Geometry` instance.

        MEOS Functions:
            minus_geo_geoset

        See Also:
            :meth:`minus`
        """
        result = minus_geo_geoset(geometry_to_gserialized(other), self._inner)
        return gserialized_to_shapely_geometry(result[0]) if result is not None else None

    def union(self, other: Union[GeoSet, shp.Geometry]) -> GeoSet:
        """
        Returns the union of ``self`` and ``other``.

        Args:
            other: A :class:`GeoSet` or :class:`Geometry` instance

        Returns:
            A :class:`GeoSet` instance.

        MEOS Functions:
            union_geoset_geo, union_set_set
        """
        if isinstance(other, shp.Geometry):
            result = union_geoset_geo(self._inner, geometry_to_gserialized(other))
            return GeoSet(_inner=result) if result is not None else None
        elif isinstance(other, GeoSet):
            result = union_set_set(self._inner, other._inner)
            return GeoSet(_inner=result) if result is not None else None
        else:
            return super().union(other)

    # ------------------------- Transformations ------------------------------------

    def round(self, max_decimals):
        return self.__class__(_inner=geoset_round(self._inner, max_decimals))


class GeometrySet(GeoSet):
    _mobilitydb_name = 'geomset'

    _parse_function = geomset_in
    _parse_value_function = lambda x: pgis_geometry_in(x, -1) if isinstance(x, str) else geometry_to_gserialized(x)


class GeographySet(GeoSet):
    _mobilitydb_name = 'geogset'


    _parse_function = geogset_in
    _parse_value_function = lambda x: pgis_geography_in(x, -1) if isinstance(x, str) else geography_to_gserialized(x)



import psycopg2
from psycopg2 import extensions, connect

from pymeos import TBool, TInt, TFloat, TText, TGeomPoint, TGeogPoint, TBox, STBox, TimestampSet, Period, PeriodSet


class MobilityDB:
    """
    Helper class to register MobilityDB classes to a psycopg2 connection and their automatic conversion to
    PyMEOS classes.
    """
    @classmethod
    def connect(cls, *args, **kwargs) -> psycopg2.extensions.connection:
        """
        Establisesh a connection to a MobilityDB server.

        Refer to :func:`psycopg2.connect` for the list of valid arguments.

        If the connection already exists, see the :func:`~pymeos.db.psycopg2.MobilityDB.register` in this class.

        Args:
            *args: positional arguments that will be passed to :func:`psycopg.connect`
            **kwargs: keyword arguments that will be passed to :func:`psycopg.connect`

        Returns:
            A new :class:`psycopg2.extensions.connection` object with the MobilityDB classes registered.

        """
        conn = connect(*args, **kwargs)
        cls.register(conn)
        return conn

    @classmethod
    def register(cls, connection: psycopg2.extensions.connection) -> None:
        """
        Registers MobilityDB classes to the passed connection.

        Args:
            connection: An :class:`psycopg2.extensions.connection` to register the classes to.
        """
        if isinstance(connection, extensions.cursor):
            # Retro compatibility
            cursor = connection
        else:
            cursor = connection.cursor()

        # Add MobilityDB types to PostgreSQL adapter and specify the reader function for each type.
        classes = [TimestampSet, Period, PeriodSet, TBox, TBool, TInt, TFloat, TText, STBox, TGeomPoint, TGeogPoint]
        for cl in classes:
            cursor.execute(f'SELECT NULL::{cl._mobilitydb_name}')
            oid = cursor.description[0][1]
            extensions.register_type(extensions.new_type((oid,), cl._mobilitydb_name, cl.read_from_cursor))



from .aggregators import *
from .boxes import *
from .main import *
from .meos_init import *
from .temporal import *
from .collections import *
from pymeos_cffi import \
    MeosException, \
    MeosInternalError, \
    MeosArgumentError, \
    MeosIoError, \
    MeosInternalTypeError, \
    MeosValueOutOfRangeError, \
    MeosDivisionByZeroError, \
    MeosMemoryAllocError, \
    MeosAggregationError, \
    MeosDirectoryError, \
    MeosFileError, \
    MeosInvalidArgError, \
    MeosInvalidArgTypeError, \
    MeosInvalidArgValueError, \
    MeosMfJsonInputError, \
    MeosMfJsonOutputError, \
    MeosTextInputError, \
    MeosTextOutputError, \
    MeosWkbInputError, \
    MeosWkbOutputError, \
    MeosGeoJsonInputError, \
    MeosGeoJsonOutputError

__version__ = '1.1.3a1'
__all__ = [
    # initialization
    'pymeos_initialize', 'pymeos_finalize',
    # boxes
    'Box', 'TBox', 'STBox',
    # main
    'TBool', 'TBoolInst', 'TBoolSeq', 'TBoolSeqSet',
    'TInt', 'TIntInst', 'TIntSeq', 'TIntSeqSet',
    'TFloat', 'TFloatInst', 'TFloatSeq', 'TFloatSeqSet',
    'TText', 'TTextInst', 'TTextSeq', 'TTextSeqSet',
    'TPointInst', 'TPointSeq', 'TPointSeqSet',
    'TGeomPoint', 'TGeomPointInst', 'TGeomPointSeq', 'TGeomPointSeqSet',
    'TGeogPoint', 'TGeogPointInst', 'TGeogPointSeq', 'TGeogPointSeqSet',
    # temporal
    'Temporal', 'TInstant', 'TSequence', 'TSequenceSet',
    # Collections
    'Time', 'Period', 'TimestampSet', 'PeriodSet',
    'TextSet',
    'IntSet', 'IntSpan', 'IntSpanSet',
    'FloatSet', 'FloatSpan', 'FloatSpanSet',
    'GeoSet', 'GeometrySet', 'GeographySet',
    # extras
    'TInterpolation',
    # aggregators
    'TemporalInstantCountAggregator', 'TemporalPeriodCountAggregator', 'TemporalExtentAggregator',
    'TemporalAndAggregator', 'TemporalOrAggregator',
    'TemporalAverageAggregator',
    'TemporalNumberExtentAggregator',
    'TemporalIntMaxAggregator', 'TemporalIntMinAggregator', 'TemporalIntSumAggregator',
    'TemporalFloatMaxAggregator', 'TemporalFloatMinAggregator', 'TemporalFloatSumAggregator',
    'TemporalTextMaxAggregator', 'TemporalTextMinAggregator',
    'TemporalPointExtentAggregator',
    'TimeInstantaneousUnionAggregator', 'TimeContinuousUnionAggregator',
    # exceptions
    'MeosException',
    'MeosInternalError',
    'MeosArgumentError',
    'MeosIoError',
    'MeosInternalTypeError',
    'MeosValueOutOfRangeError',
    'MeosDivisionByZeroError',
    'MeosMemoryAllocError',
    'MeosAggregationError',
    'MeosDirectoryError',
    'MeosFileError',
    'MeosInvalidArgError',
    'MeosInvalidArgTypeError',
    'MeosInvalidArgValueError',
    'MeosMfJsonInputError',
    'MeosMfJsonOutputError',
    'MeosTextInputError',
    'MeosTextOutputError',
    'MeosWkbInputError',
    'MeosWkbOutputError',
    'MeosGeoJsonInputError',
    'MeosGeoJsonOutputError',
]



from __future__ import annotations

from abc import ABC
from functools import reduce
from typing import Optional, Union, List, Set, overload

from pymeos_cffi import *

from ..temporal import TInterpolation, Temporal, TInstant, TSequence, TSequenceSet
from ..collections import *


class TBool(Temporal[bool, 'TBool', 'TBoolInst', 'TBoolSeq', 'TBoolSeqSet'], ABC):
    """
    Base class for temporal boolean.
    """

    _mobilitydb_name = 'tbool'

    BaseClass = bool
    _parse_function = tbool_in

    def __init__(self, _inner) -> None:
        super().__init__()

    # ------------------------- Constructors ----------------------------------
    @staticmethod
    def from_base_temporal(value: bool, base: Temporal) -> TBool:
        """
        Create a temporal Boolean from a Boolean value and the time frame of
        another temporal object.

        Args:
            value: Boolean value.
            base: Temporal object to use as time frame.

        Returns:
            A new :class:`TBool` object.

        MEOS Functions:
            tbool_from_base_temp
        """
        result = tbool_from_base_temp(value, base._inner)
        return Temporal._factory(result)

    @staticmethod
    @overload
    def from_base_time(value: bool, base: datetime) -> TBoolInst:
        ...

    @staticmethod
    @overload
    def from_base_time(value: bool, base: Union[TimestampSet, Period]) -> \
            TBoolSeq:
        ...

    @staticmethod
    @overload
    def from_base_time(value: bool, base: PeriodSet) -> TBoolSeqSet:
        ...

    @staticmethod
    def from_base_time(value: bool, base: Time) -> TBool:
        """
        Create a temporal boolean from a boolean value and a time object.

        Args:
            value: Boolean value.
            base: Time object to use as temporal dimension.

        Returns:
            A new temporal boolean.

        MEOS Functions:
            tboolinst_make, tboolseq_from_base_timestampset,
            tboolseq_from_base_period, tboolseqset_from_base_periodset
        """
        if isinstance(base, datetime):
            return TBoolInst(_inner=tboolinst_make(value,
                                                   datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TBoolSeq(_inner=tboolseq_from_base_timestampset(value,
                                                                   base._inner))
        elif isinstance(base, Period):
            return TBoolSeq(_inner=tboolseq_from_base_period(value,
                                                             base._inner))
        elif isinstance(base, PeriodSet):
            return TBoolSeqSet(_inner=tboolseqset_from_base_periodset(value,
                                                                      base._inner))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    # ------------------------- Output ----------------------------------------
    def __str__(self):
        """
        Returns the string representation of `self`.

        MEOS Function:
            tbool_out
        """
        return tbool_out(self._inner)

    def as_wkt(self):
        """
        Returns the string representation of `self` in WKT format.

        MEOS Function:
            tbool_out
        """
        return tbool_out(self._inner)

    # ------------------------- Accessors -------------------------------------
    def value_set(self) -> Set[bool]:
        """
        Returns the unique values in `self`.

        MEOS Function:
            tbool_values
        """
        values, count = tbool_values(self._inner)
        return {values[i] for i in range(count)}

    def start_value(self) -> bool:
        """
        Returns the starting value of `self`.

        MEOS Function:
            tbool_start_value
        """
        return tbool_start_value(self._inner)

    def end_value(self) -> bool:
        """
        Returns the ending value of `self`.

        MEOS Function:
            tbool_end_value
        """
        return tbool_end_value(self._inner)

    def value_at_timestamp(self, timestamp) -> bool:
        """
        Returns the value that `self` takes at a certain moment.

        Args:
            timestamp: Timestamp to get the value at.

        Returns:
            The value at the given timestamp.

        MEOS Function:
            tbool_value_at_timestamp
        """
        return tbool_value_at_timestamp(self._inner,
                                        datetime_to_timestamptz(timestamp), True)

    # ------------------------- Ever and Always Comparisons -------------------
    def always_eq(self, value: bool) -> bool:
        """
        Returns whether `self` is always equal to `value`.

        Args:
            value: Value to check for.

        Returns:
            True if `self` is always equal to `value`, False otherwise.

        MEOS Function:
            tbool_always_eq
        """
        return tbool_always_eq(self._inner, value)

    def ever_eq(self, value: bool) -> bool:
        """
        Returns whether `self` is ever equal to `value`.

        Args:
            value: Value to check for.

        Returns:
            True if `self` is ever equal to `value`, False otherwise.

        MEOS Function:
            tbool_ever_eq
        """
        return tbool_ever_eq(self._inner, value)

    def never_eq(self, value: bool) -> bool:
        """
        Returns whether `self` is never equal to `value`.

        Args:
            value: Value to check for.

        Returns:
            True if `self` is never equal to `value`, False otherwise.

        MEOS Function:
            tbool_ever_eq
        """
        return not tbool_ever_eq(self._inner, value)

    # ------------------------- Temporal Comparisons --------------------------
    def temporal_equal(self, other: Union[bool, Temporal]) -> TBool:
        """
        Returns the temporal equality relation between `self` and `other`.

        Args:
            other: A temporal or boolean object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal equality relation.

        MEOS Functions:
            teq_tbool_tbool, teq_temporal_temporal
        """
        if isinstance(other, bool):
            result = teq_tbool_bool(self._inner, other)
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[bool, Temporal]) -> TBool:
        """
        Returns the temporal inequality relation between `self` and `other`.

        Args:
            other: A temporal or boolean object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal inequality relation.

        MEOS Functions:
            tne_tbool_tbool, tne_temporal_temporal
        """
        if isinstance(other, bool):
            result = tne_tbool_bool(self._inner, other)
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    # ------------------------- Restrictions ----------------------------------
    def at(self, other: Union[bool, Time]) -> TBool:
        """
        Returns a new temporal boolean with the values of `self` restricted to
        the time or value `other`.

        Args:
            other: Time or value to restrict to.

        Returns:
            A new temporal boolean.

        MEOS Functions:
            tbool_at_value, temporal_at_timestamp, temporal_at_timestampset,
            temporal_at_period, temporal_at_periodset
        """
        if isinstance(other, bool):
            result = tbool_at_value(self._inner, other)
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[bool, Time]) -> TBool:
        """
        Returns a new temporal boolean with the values of `self` restricted to
        the complement of the time or value
         `other`.

        Args:
            other: Time or value to restrict to the complement of.

        Returns:
            A new temporal boolean.

        MEOS Functions:
            tbool_minus_value, temporal_minus_timestamp,
            temporal_minus_timestampset, temporal_minus_period,
            temporal_minus_periodset
        """
        if isinstance(other, bool):
            result = tbool_minus_value(self._inner, other)
        else:
            return super().minus(other)
        return Temporal._factory(result)

    # ------------------------- Boolean Operations ----------------------------
    def temporal_and(self, other: Union[bool, TBool]) -> TBool:
        """
        Returns the temporal conjunction of `self` and `other`.

        Args:
            other: A temporal or boolean object to combine with `self`.

        Returns:
            A :class:`TBool` with the temporal conjunction of `self` and
            `other`.

        MEOS Functions:
            tand_tbool_bool, tand_tbool_tbool
        """
        if isinstance(other, bool):
            return self.__class__(_inner=tand_tbool_bool(self._inner, other))
        elif isinstance(other, TBool):
            return self.__class__(_inner=tand_tbool_tbool(self._inner,
                                                          other._inner))
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __and__(self, other):
        """
        Returns the temporal conjunction of `self` and `other`.

        Args:
            other: A temporal or boolean object to combine with `self`.

        Returns:
            A :class:`TBool` with the temporal conjunction of `self` and `other`.

        MEOS Functions:
            tand_tbool_bool, tand_tbool_tbool
        """
        return self.temporal_and(other)

    def temporal_or(self, other: Union[bool, TBool]) -> TBool:
        """
        Returns the temporal disjunction of `self` and `other`.

        Args:
            other: A temporal or boolean object to combine with `self`.

        Returns:
            A :class:`TBool` with the temporal disjunction of `self` and `other`.

        MEOS Functions:
            tor_tbool_bool, tor_tbool_tbool
        """
        if isinstance(other, bool):
            return self.__class__(_inner=tor_tbool_bool(self._inner, other))
        elif isinstance(other, TBool):
            return self.__class__(_inner=tor_tbool_tbool(self._inner,
                                                         other._inner))
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __or__(self, other):
        """
        Returns the temporal disjunction of `self` and `other`.

        Args:
            other: A temporal or boolean object to combine with `self`.

        Returns:
            A :class:`TBool` with the temporal disjunction of `self` and `other`.

        MEOS Functions:
            tor_tbool_bool, tor_tbool_tbool
        """
        return self.temporal_or(other)

    def temporal_not(self) -> TBool:
        """
        Returns the temporal negation of `self`.

        Returns:
            A :class:`TBool` with the temporal negation of `self`.

        MEOS Function:
            tnot_tbool
        """
        return self.__class__(_inner=tnot_tbool(self._inner))

    def __neg__(self):
        """
        Returns the temporal negation of `self`.

        Returns:
            A :class:`TBool` with the temporal negation of `self`.

        MEOS Function:
            tnot_tbool
        """
        return self.temporal_not()

    def __invert__(self):
        """
        Returns the temporal negation of `self`.

        Returns:
            A :class:`TBool` with the temporal negation of `self`.

        MEOS Function:
            tnot_tbool
        """
        return self.temporal_not()

    def when_true(self) -> Optional[PeriodSet]:
        """
        Returns a period set with the periods where `self` is True.

        Returns:
            A :class:`PeriodSet` with the periods where `self` is True.

        MEOS Function:
            tbool_when_true
        """
        result = tbool_when_true(self._inner)
        return PeriodSet(_inner=result) if result is not None else None

    def when_false(self) -> Optional[PeriodSet]:
        """
        Returns a period set with the periods where `self` is False.

        Returns:
            A :class:`PeriodSet` with the periods where `self` is False.

        MEOS Function:
            tbool_when_true, tnot_tbool
        """
        result = tbool_when_true(tnot_tbool(self._inner))
        return PeriodSet(_inner=result) if result is not None else None

    # ------------------------- Database Operations ---------------------------
    @staticmethod
    def read_from_cursor(value, _=None):
        """
        Reads a :class:`TBool` from a database cursor. Used when automatically
        loading objects from the database.
        Users should use the class constructor instead.
        """
        if not value:
            return None
        if value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TBoolInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TBoolSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TBoolSeqSet(string=value)
            else:
                return TBoolSeq(string=value)
        raise Exception("ERROR: Could not parse temporal boolean value")


class TBoolInst(TInstant[bool, 'TBool', 'TBoolInst', 'TBoolSeq', 'TBoolSeqSet'], TBool):
    """
    Class for representing temporal boolean values at a single instant.
    """
    _make_function = tboolinst_make
    _cast_function = bool

    def __init__(self, string: Optional[str] = None, *,
                 value: Optional[Union[str, bool]] = None,
                 timestamp: Optional[Union[str, datetime]] = None, _inner=None):
        super().__init__(string=string, value=value, timestamp=timestamp,
                         _inner=_inner)


class TBoolSeq(TSequence[bool, 'TBool', 'TBoolInst', 'TBoolSeq',
'TBoolSeqSet'], TBool):
    """
    Class for representing temporal boolean values over a period of time.
    """
    ComponentClass = TBoolInst

    def __init__(self, string: Optional[str] = None, *,
                 instant_list: Optional[List[Union[str, TBoolInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False,
                 interpolation: TInterpolation = TInterpolation.STEPWISE,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list,
                         lower_inc=lower_inc, upper_inc=upper_inc,
                         interpolation=interpolation,
                         normalize=normalize, _inner=_inner)


class TBoolSeqSet(TSequenceSet[bool, 'TBool', 'TBoolInst', 'TBoolSeq',
'TBoolSeqSet'], TBool):
    """
    Class for representing temporal boolean values over a period of time with gaps.
    """
    ComponentClass = TBoolSeq

    def __init__(self, string: Optional[str] = None, *,
                 sequence_list: Optional[List[Union[str, TBoolSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list,
                         normalize=normalize, _inner=_inner)
